# pylint: disable=no-self-use
import json
from unittest.mock import MagicMock

import pytest

from PartSegCore.json_hooks import PartSegEncoder, partseg_object_hook
from PartSegCore.utils import (
    BaseModel,
    CallbackFun,
    CallbackMethod,
    EventedDict,
    ProfileDict,
    get_callback,
    iterate_names,
    recursive_update_dict,
)


def test_callback_fun():
    call_list = []

    def call_fun():
        call_list.append(1)

    callback = CallbackFun(call_fun)
    assert not call_list
    callback()
    assert call_list == [1]
    assert callback.is_alive()


def test_callback_method():
    call_list = []

    class A:
        def fun(self):  # pylint: disable=no-self-use
            call_list.append(1)

    a = A()
    callback = CallbackMethod(a.fun)
    assert not call_list
    callback()
    assert call_list == [1]
    assert callback.is_alive()
    del a  # skipcq: PTC-W0043
    assert not callback.is_alive()
    callback()
    assert call_list == [1]


def test_get_callback():
    def call_fun():
        raise NotImplementedError

    class A:
        def fun(self):  # pylint: disable=no-self-use
            raise NotImplementedError

    a = A()

    assert isinstance(get_callback(call_fun), CallbackFun)
    assert isinstance(get_callback(a.fun), CallbackMethod)


def test_recursive_update_dict_basic():
    dict1 = {"a": 1, "b": 2}
    dict2 = {"b": 3, "c": 4}
    dict1_copy = dict1.copy()
    dict1.update(dict2)
    recursive_update_dict(dict1_copy, dict2)
    assert dict1 == dict1_copy


def test_recursive_update_dict():
    dict1 = {"a": {"k": 1, "l": 2}, "b": {"k": 1, "l": 2}}
    dict2 = {"a": {"m": 3, "l": 4}, "b": 3, "c": 4}
    recursive_update_dict(dict1, dict2)
    assert dict1 == {"a": {"k": 1, "l": 4, "m": 3}, "b": 3, "c": 4}


class TestEventedDict:
    def test_simple_add(self):
        receiver = MagicMock()

        dkt = EventedDict()
        dkt.setted.connect(receiver.set)
        dkt.deleted.connect(receiver.delete)
        dkt["a"] = 1
        assert dkt["a"] == 1
        assert receiver.set.call_count == 1
        assert "'a': 1" in str(dkt)
        assert "'a': 1" in repr(dkt)
        dkt["a"] = 2
        assert dkt["a"] == 2
        assert receiver.set.call_count == 2
        assert len(dkt) == 1
        del dkt["a"]
        assert receiver.set.call_count == 2
        assert receiver.delete.call_count == 1
        assert len(dkt) == 0

    def test_simple_add_remove(self):
        callback_list = []

        def callback_add():
            callback_list.append(1)

        def callback_delete():
            callback_list.append(2)

        dkt = EventedDict()
        dkt.setted.connect(callback_add)
        dkt.deleted.connect(callback_delete)

        dkt[1] = 1
        dkt[2] = 1
        assert len(dkt) == 2
        assert callback_list == [1, 1]
        del dkt[1]
        assert len(dkt) == 1
        assert callback_list == [1, 1, 2]

    def test_nested_evented(self):
        dkt = EventedDict(bar={"foo": {"baz": 1}})
        assert isinstance(dkt["bar"], EventedDict)
        assert isinstance(dkt["bar"]["foo"], EventedDict)
        assert dkt["bar"]["foo"]["baz"] == 1

        dkt["baz"] = {"bar": {"foo": 1}}
        assert isinstance(dkt["baz"], EventedDict)
        assert isinstance(dkt["baz"]["bar"], EventedDict)
        assert dkt["baz"]["bar"]["foo"] == 1

    def test_serialize(self, tmp_path):
        dkt = EventedDict(a={"b": {"c": 1, "d": 2, "e": 3}, "f": 1}, g={"h": {"i": 1, "j": 2}, "k": [6, 7, 8]})
        with (tmp_path / "test_dict.json").open("w") as f_p:
            json.dump(dkt, f_p, cls=PartSegEncoder)

        with (tmp_path / "test_dict.json").open("r") as f_p:
            dkt2 = json.load(f_p, object_hook=partseg_object_hook)

        assert isinstance(dkt2, EventedDict)
        assert isinstance(dkt2["a"], EventedDict)
        assert dkt["g"]["k"] == [6, 7, 8]

    def test_signal_names(self):
        receiver = MagicMock()
        dkt = EventedDict(baz={"foo": 1})
        dkt.setted.connect(receiver.set)
        dkt.deleted.connect(receiver.deleted)
        dkt["foo"] = 1
        assert receiver.set.call_count == 1
        receiver.set.assert_called_with("foo")
        dkt["bar"] = EventedDict()
        assert receiver.set.call_count == 2
        receiver.set.assert_called_with("bar")
        dkt["bar"]["baz"] = 1
        assert receiver.set.call_count == 3
        receiver.set.assert_called_with("bar.baz")
        dkt["baz"]["foo"] = 2
        assert receiver.set.call_count == 4
        receiver.set.assert_called_with("baz.foo")

        del dkt["bar"]["baz"]
        assert receiver.deleted.call_count == 1
        receiver.deleted.assert_called_with("bar.baz")
        del dkt["bar"]
        assert receiver.deleted.call_count == 2
        receiver.deleted.assert_called_with("bar")

    def test_propagate_signal(self):
        receiver = MagicMock()
        dkt = EventedDict(baz={"foo": 1})
        dkt.setted.connect(receiver.set)
        dkt.deleted.connect(receiver.deleted)
        dkt["baz"].base_key = ""
        dkt["baz"]["foo"] = 2
        receiver.set.assert_called_with("foo")
        receiver.set.assert_called_once()
        del dkt["baz"]["foo"]
        receiver.deleted.assert_called_with("foo")
        receiver.deleted.assert_called_once()

    def test_dict_force_class(self):
        dkt = EventedDict(int)
        dkt["a"] = 1
        with pytest.raises(TypeError):
            dkt["b"] = "a"
        assert dkt["a"] == 1
        assert set(dkt) == {"a"}

    def test_dict_force_class_dkt(self):
        dkt = EventedDict({"a": int, "b": tuple})
        dkt["a"] = 1
        with pytest.raises(TypeError):
            dkt["b"] = "a"
        dkt["b"] = (1, 2)
        dkt["c"] = "A"
        assert set(dkt) == {"a", "b", "c"}

    def test_dict_force_class_dkt2(self):
        dkt = EventedDict({"a": int, "b": tuple, "*": str})
        dkt["a"] = 1
        with pytest.raises(TypeError):
            dkt["b"] = "a"
        dkt["b"] = (1, 2)
        with pytest.raises(TypeError):
            dkt["c"] = 1
        dkt["c"] = "A"
        assert set(dkt) == {"a", "b", "c"}

    def test_dict_nested_type(self):
        dkt = EventedDict({"a": {"b": int}})
        dkt["a"] = {"b": 1}
        with pytest.raises(TypeError):
            dkt["a"] = {"b": "a"}
        assert dkt["a"]["b"] == 1
        assert set(dkt) == {"a"}
        assert set(dkt["a"]) == {"b"}

    def test_dict_repr(self):
        assert repr(EventedDict(a=1, b=2)) == "EventedDict(klass={'*': <class 'object'>}, {'a': 1, 'b': 2})"
        assert str(EventedDict(a=1, b=2)) == "EventedDict[{'*': <class 'object'>}]({'a': 1, 'b': 2})"


class TestProfileDict:
    def test_simple(self):
        dkt = ProfileDict()
        dkt.set("a.b.c", 1)
        dkt.set("a.b.a", 2)
        assert dkt.get("a.b.c") == 1
        with pytest.raises(KeyError):
            dkt.get("a.b.d")
        dkt.get("a.b.d", 3)
        assert dkt.get("a.b.d") == 3
        assert dkt.get("a.b") == {"a": 2, "c": 1, "d": 3}
        with pytest.raises(TypeError):
            dkt.set("a.b.c.d", 3)

    def test_update(self):
        dkt = ProfileDict()
        dkt.update(a=1, b=2, c=3)
        assert dkt.my_dict == {"a": 1, "b": 2, "c": 3}
        dkt2 = ProfileDict()
        dkt2.update(c=4, d={"a": 2, "e": 7})
        assert dkt2.get("d.e") == 7
        dkt.update(dkt2)
        assert dkt.get("d.e") == 7
        assert dkt.get("c") == 4
        dkt.update({"g": 1, "h": 4})
        assert dkt.get("g") == 1
        dkt.update({"w": 1, "z": 4}, w=3)
        assert dkt.get("w") == 3
        assert dkt.verify_data()
        assert dkt.pop_errors() == []
        dkt.set("e.h.l", {"aaa": 1, "__error__": True})
        assert not dkt.verify_data()
        assert dkt.pop_errors()[0][0] == "e.h"

    def test_serialize(self, tmp_path):
        dkt = ProfileDict()
        dkt.set("a.b.c", 1)
        dkt.set("a.b.a", 2)
        with open(tmp_path / "test.json", "w") as f_p:
            json.dump(dkt, f_p, cls=PartSegEncoder)
        with open(tmp_path / "test.json") as f_p:
            dkt2 = json.load(f_p, object_hook=partseg_object_hook)

        assert dkt.my_dict == dkt2.my_dict

    def test_callback(self):
        def dummy_call():
            receiver.dummy()

        receiver = MagicMock()

        dkt = ProfileDict()
        dkt.connect("", receiver.empty)
        dkt.connect("", dummy_call)
        dkt.connect("b", receiver.b)
        dkt.connect(["d", "c"], receiver.dc)

        dkt.set("test.a", 1)
        assert receiver.empty.call_count == 1
        assert receiver.dummy.call_count == 1
        receiver.empty.assert_called_with("a")
        receiver.dummy.assert_called_with()
        dkt.set("test.a", 1)
        assert receiver.empty.call_count == 1
        receiver.b.assert_not_called()
        dkt.set("test2.a", 1)
        assert receiver.empty.call_count == 2
        receiver.b.assert_not_called()
        dkt.set(["test", "b"], 1)
        assert receiver.empty.call_count == 3
        assert receiver.b.call_count == 1
        dkt.set("test.d.c", 1)
        receiver.dc.assert_called_once()
        dkt.set("test.a", 2)
        assert receiver.empty.call_count == 5

    def test_dict_add(self):
        dkt = ProfileDict()
        data = dkt.get("foo", default={})
        assert isinstance(data, EventedDict)
        dkt.update({"foo": {"bar": 1}})
        data["a"] = 1
        assert dkt.get("foo") == {"a": 1, "bar": 1}

    def test_force_type(self):
        dkt = ProfileDict(int)
        dkt.set("a", 1)
        with pytest.raises(TypeError):
            dkt.set("b", "a")

    def test_force_type_nested(self):
        dkt = ProfileDict({"a": int, "b": {"a": int, "*": str}})
        dkt.set("a", 1)
        with pytest.raises(TypeError):
            dkt.set("b", "a")
        dkt.set("b.a", 1)
        with pytest.raises(TypeError):
            dkt.set("b.c", 1)
        with pytest.raises(TypeError):
            dkt.set("a", "a")
        assert dkt.get("a") == 1
        assert dkt.get("b") == {"a": 1}


def test_iterate_names():
    assert iterate_names("aaaa", {}) == "aaaa"
    assert iterate_names("aaaa", {"aaaa"}) == "aaaa (1)"
    assert iterate_names("aaaa", {"aaaa", "aaaa (1)", "aaaa (3)"}) == "aaaa (2)"

    assert iterate_names("a" * 10, {}, 10) == "a" * 10
    assert iterate_names("a" * 11, {}, 10) == "a" * 10
    assert iterate_names("a" * 9, {"a" * 9}, 10) == "a" * 5 + " (1)"

    input_set = {"aaaaa"}
    for _ in range(15):
        input_set.add(iterate_names("a" * 5, input_set))
    assert iterate_names("a" * 5, input_set) == "a" * 5 + " (16)"
    for _ in range(85):
        input_set.add(iterate_names("a" * 5, input_set))
    assert iterate_names("a" * 5, input_set) is None


def test_base_model_getitem():
    class SampleModel(BaseModel):
        a: int = 1
        b: float = 2.0
        c: str = "3"

    ob = SampleModel()
    with pytest.warns(FutureWarning, match=r"Access to attribute by \[\] is deprecated\. Use \. instead"):
        assert ob["a"] == 1
    with pytest.warns(FutureWarning, match=r"Access to attribute by \[\] is deprecated\. Use \. instead"):
        assert ob["b"] == 2.0
    with pytest.warns(FutureWarning, match=r"Access to attribute by \[\] is deprecated\. Use \. instead"):
        assert ob["c"] == "3"
    with pytest.raises(KeyError):
        ob["d"]  # pylint: disable=pointless-statement



import local_migrator

from PartSegCore._old_json_hooks import part_hook

PLUGINS_STR = "plugins."
PLUGINS_STR_LEN = len(PLUGINS_STR)


class PartSegEncoder(local_migrator.Encoder):
    pass


def _validate_plugin_prefix(dkt: dict):
    if dkt["__class__"].startswith(PLUGINS_STR):
        # workaround for plans exported from an old PartSeg bundle
        dkt["__class__"] = dkt["__class__"][PLUGINS_STR_LEN:]
        if "__class_version_dkt__" in dkt:
            for name, value in list(dkt["__class_version_dkt__"].items()):
                if name.startswith(PLUGINS_STR):
                    dkt["__class_version_dkt__"][name[PLUGINS_STR_LEN:]] = value
                    del dkt["__class_version_dkt__"][name]


def partseg_object_hook(dkt: dict):
    if "__class__" in dkt:
        _validate_plugin_prefix(dkt)
        return local_migrator.object_hook(dkt)

    if "__ReadOnly__" in dkt or "__Serializable__" in dkt or "__Enum__" in dkt:
        if problematic_fields := local_migrator.check_for_errors_in_dkt_values(dkt):
            dkt["__error__"] = f"Error in fields: {', '.join(problematic_fields)}"
            return dkt
        is_enum = "__Enum__" in dkt
        for el in ("__Enum__", "__Serializable__", "__ReadOnly__"):
            dkt.pop(el, None)
        cls_str = dkt["__subtype__"]
        del dkt["__subtype__"]
        try:
            dkt_migrated = local_migrator.REGISTER.migrate_data(cls_str, {}, dkt)
            cls = local_migrator.REGISTER.get_class(cls_str)
            return cls(**dkt_migrated)
        except Exception:  # pylint: disable=broad-except
            dkt["__subtype__"] = cls_str
            dkt["__Enum__" if is_enum else "__Serializable__"] = True

    return part_hook(dkt)



from enum import Enum

from local_migrator import register_class


# noinspection NonAsciiCharacters
@register_class(old_paths=["PartSeg.utils.universal_const.Units"])
class Units(Enum):
    mm = 0
    µm = 1  # noqa: PLC2401
    nm = 2
    pm = 3

    def __str__(self):
        return self.name.replace("_", " ")


_UNITS_LIST = ["mm", "µm", "nm", "pm"]
UNIT_SCALE = [10**3, 10**6, 10**9, 10**12]


@register_class()
class LayerNamingFormat(Enum):
    channel_only = 0
    filename_only = 1
    filename_channel = 2
    channel_filename = 3

    def __str__(self):
        return self.name.replace("_", " ")


def format_layer_name(layer_format: LayerNamingFormat, file_name: str, channel_name: str) -> str:
    if layer_format == LayerNamingFormat.channel_only:
        return channel_name
    if layer_format == LayerNamingFormat.filename_only:
        return file_name
    if layer_format == LayerNamingFormat.filename_channel:
        return f"{file_name} | {channel_name}"
    if layer_format == LayerNamingFormat.channel_filename:
        return f"{channel_name} | {file_name}"
    raise ValueError("Unknown format")  # pragma: no cover



# pylint: disable=no-self-use

import numpy as np

from PartSegCore import UNIT_SCALE, Units
from PartSegCore.mask_partition_utils import BorderRim, MaskDistanceSplit


class TestBorderRim:
    def test_base_2d(self):
        mask = np.zeros((6, 6), dtype=np.uint8)
        mask[1:-1, 1:-1] = 1
        nm_scalar = UNIT_SCALE[Units.nm.value]
        voxel_size = (1 / nm_scalar,) * 2
        result_mask = BorderRim.border_mask(mask, 1, Units.nm, voxel_size)
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2] = 0
        assert np.all(mask2 == result_mask)

    def test_scaling_2d(self):
        mask = np.zeros((6, 12), dtype=np.uint8)
        mask[1:-1, 2:-2] = 1
        nm_scalar = UNIT_SCALE[Units.nm.value]
        voxel_size = (2 / nm_scalar, 1 / nm_scalar)
        result_mask = BorderRim.border_mask(mask, 2, Units.nm, voxel_size)
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4] = 0
        assert np.all(mask2 == result_mask)

    def test_base_3d(self):
        mask = np.zeros((20, 20, 20), dtype=np.uint8)
        mask[2:-2, 2:-2, 2:-2] = 1
        mask2 = np.copy(mask)
        mask2[3:-3, 3:-3, 3:-3] = 0
        nm_scalar = UNIT_SCALE[Units.nm.value]
        voxel_size = (1 / nm_scalar,) * 3
        result_mask = BorderRim.border_mask(mask, 1, Units.nm, voxel_size)
        assert np.all(result_mask == mask2)

    def test_scaling_3d(self):
        mask = np.zeros((10, 20, 20), dtype=np.uint8)
        mask[1:-1, 2:-2, 2:-2] = 1
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4, 4:-4] = 0
        nm_scalar = UNIT_SCALE[Units.nm.value]
        voxel_size = (2 / nm_scalar, 1 / nm_scalar, 1 / nm_scalar)
        result_mask = BorderRim.border_mask(mask, 2, Units.nm, voxel_size)
        assert np.all(result_mask == mask2)


class TestSplitMaskOnPart:
    def test_base_2d_thick(self):
        mask = np.zeros((6, 6), dtype=np.uint8)
        mask[1:-1, 1:-1] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, False, (1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, False, (1, 1))
        assert np.all(mask2 == result_mask)

    def test_scaling_2d_thick(self):
        mask = np.zeros((6, 12), dtype=np.uint8)
        mask[1:-1, 2:-2] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, False, (2, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, False, (2, 1))
        assert np.all(mask2 == result_mask)

    def test_base_3d_thick(self):
        mask = np.zeros((6, 6, 6), dtype=np.uint8)
        mask[1:-1, 1:-1, 1:-1] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, False, (1, 1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2, 2:-2] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, False, (1, 1, 1))
        assert np.all(mask2 == result_mask)

    def test_scaling_3d_thick(self):
        mask = np.zeros((6, 12, 12), dtype=np.uint8)
        mask[1:-1, 2:-2, 2:-2] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, False, (2, 1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4, 4:-4] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, False, (2, 1, 1))
        assert np.all(mask2 == result_mask)

    def test_base_2d_volume(self):
        mask = np.zeros((7, 7), dtype=np.uint8)
        mask[1:-1, 1:-1] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, True, (1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, True, (1, 1))
        assert np.all(mask2 == result_mask)

    def test_scaling_2d_volume(self):
        mask = np.zeros((7, 14), dtype=np.uint8)
        mask[1:-1, 2:-2] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, True, (2, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, True, (2, 1))
        assert np.all(mask2 == result_mask)

    def test_base_3d_volume(self):
        mask = np.zeros((7, 7, 7), dtype=np.uint8)
        mask[1:-1, 1:-1, 1:-1] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, True, (1, 1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2, 2:-2] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, True, (1, 1, 1))
        assert np.all(mask2 == result_mask)

    def test_scaling_3d_volume(self):
        mask = np.zeros((7, 14, 14), dtype=np.uint8)
        mask[1:-1, 2:-2, 2:-2] = 1
        result_mask = MaskDistanceSplit.split(mask, 1, True, (2, 1, 1))
        assert np.all(mask == result_mask)
        mask2 = np.copy(mask)
        mask2[2:-2, 4:-4, 4:-4] = 2
        result_mask = MaskDistanceSplit.split(mask, 2, True, (2, 1, 1))
        assert np.all(mask2 == result_mask)



import locale
import os
from enum import Enum

from qtpy.QtCore import Qt
from qtpy.QtGui import QKeyEvent, QResizeEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QBoxLayout,
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)
from superqt import QEnumComboBox

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_gui.searchable_combo_box import SearchComboBox
from PartSeg.common_gui.universal_gui_part import ChannelComboBox
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore.analysis.measurement_calculation import FILE_NAME_STR, MeasurementProfile, MeasurementResult
from PartSegCore.universal_const import Units

NO_MEASUREMENT_STRING = "<none>"


class FileNamesEnum(Enum):
    No = 1
    Short = 2
    Full = 3

    def __str__(self):
        return self.name


class MeasurementsStorage:
    """class for storage measurements result"""

    def __init__(self):
        self.header = []
        self.max_rows = 0
        self.content = []
        self.measurements = []
        self.expand = True
        self.show_units = False

    def clear(self):
        """clear storage"""
        self.header = []
        self.max_rows = 0
        self.content = []
        self.measurements: list[MeasurementResult] = []

    def get_size(self, save_orientation: bool):
        if save_orientation:
            return self.max_rows, len(self.header)

        return len(self.header), self.max_rows

    def set_expand(self, expand):
        if self.expand != expand:
            self.expand = expand
            self.refresh()

    def set_show_units(self, show_units):
        if self.show_units != show_units:
            self.show_units = show_units
            self.refresh()

    def refresh(self):
        self.header = []
        self.content = []
        self.max_rows = 0
        previous_labels = []
        previous_units = []
        for data in self.measurements:
            labels = data.get_labels(self.expand)
            if labels != previous_labels:
                self.content.append(labels)
                self.header.append("Name")
                previous_labels = labels
            if self.expand:
                values = data.get_separated()
                units = data.get_units()
            else:
                values, units = zip(*list(data.values()))
                values = [values]
            self.max_rows = max(self.max_rows, len(values[0]))
            self.content.extend(values)
            self.header.extend(["Value" for _ in range(len(values))])
            if self.show_units and units != previous_units:
                previous_units = units
                self.content.append(previous_units)
                self.header.append("Units")

    def add_measurements(self, data: MeasurementResult):
        self.measurements.append(data)
        self.refresh()

    def get_val_as_str(self, x: int, y: int, save_orientation: bool) -> str:
        """get value from given index"""
        if not save_orientation:
            x, y = y, x
        if len(self.content) <= x:
            return ""
        sublist = self.content[x]
        if len(sublist) <= y:
            return ""
        val = sublist[y]
        return locale.str(val) if isinstance(val, float) else str(val)

    def get_header(self, save_orientation: bool) -> list[str]:
        if save_orientation:
            return [str(i) for i in range(self.max_rows)]

        return self.header

    def get_rows(self, save_orientation: bool) -> list[str]:
        return self.get_header(not save_orientation)


class MeasurementWidgetBase(QWidget):
    """
    :type settings: Settings
    :type segment: Segment
    """

    def __init__(self, settings: PartSettings, segment=None):  # noqa: PLR0915
        super().__init__()
        self.settings = settings
        self.segment = segment
        self.measurements_storage = MeasurementsStorage()
        self.recalculate_button = QPushButton("Recalculate and\n replace measurement", self)
        self.recalculate_button.clicked.connect(self.replace_measurement_result)
        self.recalculate_append_button = QPushButton("Recalculate and\n append measurement", self)
        self.recalculate_append_button.clicked.connect(self.append_measurement_result)
        self.copy_button = QPushButton("Copy to clipboard", self)
        self.copy_button.setToolTip("You can copy also with 'Ctrl+C'. To get raw copy copy with 'Ctrl+Shit+C'")
        self.horizontal_measurement_present = QCheckBox("Horizontal view", self)
        self.no_header = QCheckBox("No header", self)
        self.no_units = QCheckBox("No units", self)
        self.no_units.setChecked(True)
        self.no_units.clicked.connect(self.refresh_view)
        self.expand_mode = QCheckBox("Expand", self)
        self.expand_mode.setChecked(True)
        self.expand_mode.setToolTip("Shows results for each component in separate entry")
        self.file_names = QEnumComboBox(enum_class=FileNamesEnum)
        self.file_names_label = QLabel("Add file name:")
        self.file_names.currentIndexChanged.connect(self.refresh_view)
        self.horizontal_measurement_present.stateChanged.connect(self.refresh_view)
        self.expand_mode.stateChanged.connect(self.refresh_view)
        self.copy_button.clicked.connect(self.copy_to_clipboard)
        self.measurement_type = SearchComboBox(self)
        # noinspection PyUnresolvedReferences
        self.measurement_type.currentIndexChanged.connect(self.measurement_profile_selection_changed)
        self.measurement_type.setToolTip(
            'You can create new measurement profile in advanced window, in tab "Measurement settings"'
        )
        self.units_choose = QEnumComboBox(enum_class=Units)
        self.units_choose.setCurrentEnum(self.settings.get("units_value", Units.nm))
        self.settings.measurement_profiles_changed.connect(self.update_measurement_list)
        self.info_field = QTableWidget(self)
        self.info_field.setColumnCount(3)
        self.info_field.setHorizontalHeaderLabels(["Name", "Value", "Units"])
        self.measurement_add_shift = 0
        layout = QVBoxLayout()
        v_butt_layout = QVBoxLayout()
        v_butt_layout.setSpacing(1)
        self.up_butt_layout = QHBoxLayout()
        self.up_butt_layout.addWidget(self.recalculate_button)
        self.up_butt_layout.addWidget(self.recalculate_append_button)
        self.butt_layout = QHBoxLayout()
        self.butt_layout.addWidget(self.horizontal_measurement_present, 1)
        self.butt_layout.addWidget(self.no_header, 1)
        self.butt_layout.addWidget(self.no_units, 1)
        self.butt_layout.addWidget(self.expand_mode, 1)
        self.butt_layout.addWidget(self.file_names_label)
        self.butt_layout.addWidget(self.file_names, 1)
        self.butt_layout.addWidget(self.copy_button, 2)
        self.butt_layout2 = QHBoxLayout()
        self.butt_layout3 = QHBoxLayout()
        self.butt_layout3.addWidget(QLabel("Units:"))
        self.butt_layout3.addWidget(self.units_choose)
        self.butt_layout3.addWidget(QLabel("Measurement set:"))
        self.butt_layout3.addWidget(self.measurement_type, 2)
        v_butt_layout.addLayout(self.up_butt_layout)
        v_butt_layout.addLayout(self.butt_layout)
        v_butt_layout.addLayout(self.butt_layout2)
        v_butt_layout.addLayout(self.butt_layout3)
        layout.addLayout(v_butt_layout)
        layout.addWidget(self.info_field)
        self.setLayout(layout)
        # noinspection PyArgumentList
        self.clip = QApplication.clipboard()
        self.previous_profile = None
        self.update_measurement_list()

    def check_if_measurement_can_be_calculated(self, name):  # pragma: no cover
        raise NotImplementedError

    def _get_mask(self):
        raise NotImplementedError

    def measurement_profile_selection_changed(self, index):
        text = self.measurement_type.itemText(index)
        text = self.check_if_measurement_can_be_calculated(text)
        try:
            stat = self.settings.measurement_profiles[text]
            is_mask = stat.is_any_mask_measurement()
            disable = is_mask and (self._get_mask() is None)
        except KeyError:
            disable = True
        self.recalculate_button.setDisabled(disable)
        self.recalculate_append_button.setDisabled(disable)
        if disable:
            self.recalculate_button.setToolTip("Measurement profile contains mask measurements when mask is not loaded")
            self.recalculate_append_button.setToolTip(
                "Measurement profile contains mask measurements when mask is not loaded"
            )
        else:
            self.recalculate_button.setToolTip("")
            self.recalculate_append_button.setToolTip("")

    def copy_to_clipboard(self):
        s = ""
        for r in range(self.info_field.rowCount()):
            for c in range(self.info_field.columnCount()):
                try:
                    s += str(self.info_field.item(r, c).text()) + "\t"
                except AttributeError:  # noqa: PERF203
                    s += "\t"
            s = s[:-1] + "\n"  # eliminate last '\t'
        self.clip.setText(s)

    def replace_measurement_result(self):
        self.measurements_storage.clear()
        self.previous_profile = ""
        self.append_measurement_result()

    def _update_file_info(self, save_orientation, rows, columns):
        if self.info_field.item(0, 0).text() != FILE_NAME_STR:
            return
        if self.file_names.currentEnum() == FileNamesEnum.No:
            if save_orientation:
                self.info_field.removeColumn(0)
            else:
                self.info_field.removeRow(0)
        elif self.file_names.currentEnum() == FileNamesEnum.Short:
            if save_orientation:
                columns = 1
            else:
                rows = 1
            for x in range(rows):
                for y in range(columns):
                    item = self.info_field.item(x, y)
                    item.setText(os.path.basename(item.text()))

    def refresh_view(self):
        self.measurements_storage.set_expand(self.expand_mode.isChecked())
        self.measurements_storage.set_show_units(not self.no_units.isChecked())
        self.info_field.clear()
        save_orientation = self.horizontal_measurement_present.isChecked()
        columns, rows = self.measurements_storage.get_size(save_orientation)
        self.info_field.setColumnCount(columns)
        self.info_field.setRowCount(rows)
        self.info_field.setHorizontalHeaderLabels(self.measurements_storage.get_header(save_orientation))
        self.info_field.setVerticalHeaderLabels(self.measurements_storage.get_rows(save_orientation))
        for x in range(rows):
            for y in range(columns):
                self.info_field.setItem(
                    x, y, QTableWidgetItem(self.measurements_storage.get_val_as_str(x, y, save_orientation))
                )
        self._update_file_info(save_orientation, rows, columns)
        self.info_field.setEditTriggers(QAbstractItemView.NoEditTriggers)

    def append_measurement_result(self):  # pragma: no cover
        raise NotImplementedError

    def keyPressEvent(self, e: QKeyEvent):
        if not e.modifiers() & Qt.KeyboardModifier.ControlModifier:
            return
        selected = self.info_field.selectedRanges()

        if e.key() == Qt.Key.Key_C:  # copy
            s = ""

            for r in range(selected[0].topRow(), selected[0].bottomRow() + 1):
                for c in range(selected[0].leftColumn(), selected[0].rightColumn() + 1):
                    try:
                        s += str(self.info_field.item(r, c).text()) + "\t"
                    except AttributeError:  # noqa: PERF203
                        s += "\t"
                s = s[:-1] + "\n"  # eliminate last '\t'
            self.clip.setText(s)

    def update_measurement_list(self):
        self.measurement_type.blockSignals(True)
        available = sorted(self.settings.measurement_profiles.keys())
        text = self.measurement_type.currentText()
        try:
            index = available.index(text) + 1
        except ValueError:
            index = 0
        self.measurement_type.clear()
        self.measurement_type.addItem(NO_MEASUREMENT_STRING)
        self.measurement_type.addItems(available)
        self.measurement_type.setCurrentIndex(index)
        self.measurement_type.blockSignals(False)

    @staticmethod
    def _move_widgets(widgets_list: list[tuple[QWidget, int]], layout1: QBoxLayout, layout2: QBoxLayout):
        for el in widgets_list:
            layout1.removeWidget(el[0])
            layout2.addWidget(el[0], el[1])

    def resizeEvent(self, _event: QResizeEvent) -> None:
        if self.width() < 800 and self.butt_layout2.count() == 0:
            self._move_widgets(
                [(self.file_names_label, 1), (self.file_names, 1), (self.copy_button, 2)],
                self.butt_layout,
                self.butt_layout2,
            )
        elif self.width() > 800 and self.butt_layout2.count() != 0:
            self._move_widgets(
                [(self.file_names_label, 1), (self.file_names, 1), (self.copy_button, 2)],
                self.butt_layout2,
                self.butt_layout,
            )


class MeasurementWidget(MeasurementWidgetBase):
    """
    :type settings: Settings
    :type segment: Segment
    """

    def __init__(self, settings: PartSettings, segment=None):
        super().__init__(settings, segment)
        self.channels_chose = ChannelComboBox()
        self.butt_layout3.insertWidget(0, QLabel("Channel:"))
        self.butt_layout3.insertWidget(1, self.channels_chose)
        self.settings.image_channel_count_changed.connect(self.image_changed)

    def _get_mask(self):
        return self.settings.mask

    def append_measurement_result(self):
        try:
            compute_class = self.settings.measurement_profiles[self.measurement_type.currentText()]
        except KeyError:
            QMessageBox.warning(
                self,
                "Measurement profile not found",
                f"Measurement profile '{self.measurement_type.currentText()}' not found",
            )
            return

        if self.settings.roi is None:
            return
        units = self.units_choose.currentEnum()

        for num in compute_class.get_channels_num():
            if not self.settings.image.has_channel(num):
                QMessageBox.warning(
                    self,
                    "Measurement error",
                    f"Cannot calculate this measurement because image do not have channel {num}",
                )
                return

        dial = ExecuteFunctionDialog(
            compute_class.calculate,
            [self.settings.image, self.channels_chose.currentIndex(), self.settings.roi_info, units],
            text="Measurement calculation",
        )  # , exception_hook=exception_hook)
        dial.exec_()
        stat: MeasurementResult = dial.get_result()
        if stat is None:
            return
        stat.set_filename(self.settings.image_path)
        self.measurements_storage.add_measurements(stat)
        self.previous_profile = compute_class.name
        self.refresh_view()

    def image_changed(self, channels_num):
        self.channels_chose.change_channels_num(channels_num)

    def check_if_measurement_can_be_calculated(self, name):
        if name in (NO_MEASUREMENT_STRING, ""):
            return NO_MEASUREMENT_STRING
        profile: MeasurementProfile = self.settings.measurement_profiles.get(name)
        if profile.is_any_mask_measurement() and self.settings.mask is None:
            QMessageBox.information(
                self, "Need mask", "To use this measurement set please use data with mask loaded", QMessageBox.Ok
            )
            self.measurement_type.setCurrentIndex(0)
            return NO_MEASUREMENT_STRING
        if self.settings.roi is None:
            QMessageBox.information(
                self,
                "Need segmentation",
                'Before calculating please create segmentation ("Execute" button)',
                QMessageBox.Ok,
            )
            self.measurement_type.setCurrentIndex(0)
            return NO_MEASUREMENT_STRING
        return name



from math import log
from typing import Union

import numpy as np
from napari.utils import Colormap
from napari.utils.colormaps import make_colorbar
from qtpy import QtGui
from qtpy.QtCore import QRect, QSize
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QLabel

from PartSeg.common_backend.base_settings import ViewSettings
from PartSeg.common_gui.napari_image_view import ImageView
from PartSeg.common_gui.numpy_qimage import NumpyQImage

canvas_icon_size = QSize(20, 20)
step = 1.01
max_step = log(1.2, step)


class ColorBar(QLabel):
    def __init__(self, settings: ViewSettings, image_view: Union[list[ImageView], ImageView]):
        super().__init__()
        self.image_view = image_view
        self._settings = settings
        self.image = None
        if isinstance(image_view, list):
            for el in image_view:
                el.channel_control.change_channel.connect(self.update_colormap)
        else:
            image_view.channel_control.change_channel.connect(self.update_colormap)
        self.range = None
        self.round_range = None
        self.setFixedWidth(80)

    def update_colormap(self, name, channel_id):
        fixed_range = self._settings.get_from_profile(f"{name}.lock_{channel_id}", False)
        gamma = self._settings.get_from_profile(f"{name}.gamma_value_{channel_id}", 1)
        if fixed_range:
            self.range = self._settings.get_from_profile(f"{name}.range_{channel_id}")
        else:
            self.range = self._settings.border_val[channel_id]
        cmap = self._settings.colormap_dict[self._settings.get_channel_colormap_name(name, channel_id)][0]

        round_factor = self.round_base(self.range[1])
        self.round_range = (
            int(round(self.range[0] / round_factor) * round_factor),
            int(round(self.range[1] / round_factor) * round_factor),
        )
        if self.round_range[0] < self.range[0]:
            self.round_range = self.round_range[0] + round_factor, self.round_range[1]
        if self.round_range[1] > self.range[1]:
            self.round_range = self.round_range[0], self.round_range[1] - round_factor
        data = np.linspace(0, 1, 512)
        interpolated = cmap.map(data)
        data = data**gamma
        colormap = Colormap(interpolated, controls=data)
        self.image = NumpyQImage(np.array(make_colorbar(colormap, size=(512, 1), horizontal=False)[::-1]))
        self.repaint()

    @staticmethod
    def round_base(val):
        if val > 10000:
            return 1000
        if val > 1000:
            return 100
        return 10 if val > 100 else 1

    @staticmethod
    def number_of_marks(val):
        if val < 500:
            return 6
        return 21 if val > 1300 else 11

    def paintEvent(self, event: QtGui.QPaintEvent):
        bar_width = 30

        if self.image is None:
            return

        rect = self.rect()
        number_of_marks = self.number_of_marks(rect.height())
        image_rect = QRect(rect.topLeft(), QSize(bar_width, rect.size().height()))
        painter = QPainter(self)
        old_font = painter.font()
        new_font = painter.font()
        new_font.setPointSizeF(new_font.pointSizeF() / 1.1)
        painter.setFont(new_font)
        painter.drawImage(image_rect, self.image)
        if self.range[1] == self.range[0]:
            painter.drawText(bar_width + 5, 20, f"{self.range[1]}")
            painter.drawText(bar_width + 5, rect.size().height(), f"{self.range[1]}")
            painter.setFont(old_font)
            return
        start_prop = 1 - (self.round_range[0] - self.range[0]) / (self.range[1] - self.range[0])
        end_prop = 1 - (self.round_range[1] - self.range[0]) / (self.range[1] - self.range[0])
        for pos, val in zip(
            np.linspace(10 + end_prop * rect.size().height(), start_prop * rect.size().height(), number_of_marks),
            np.linspace(self.round_range[1], self.round_range[0], number_of_marks, dtype=np.uint32),
        ):
            painter.drawText(bar_width + 5, int(pos), f"{val}")
        painter.setFont(old_font)



import os

from magicgui.widgets import Container, create_widget
from napari.layers import Layer

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_backend import napari_get_settings
from PartSegCore import Units
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.universal_const import LayerNamingFormat

_SETTINGS = None


class PartSegNapariEncoder(PartSegEncoder):
    def default(self, o):
        if isinstance(o, Layer):
            return o.name
        return super().default(o)


class PartSegNapariSettings(PartSettings):
    json_encoder_class = PartSegNapariEncoder

    @property
    def io_units(self) -> Units:
        return self.get("io_units", Units.nm)

    @io_units.setter
    def io_units(self, value: Units):
        self.set("io_units", value)

    @property
    def layer_naming_format(self) -> LayerNamingFormat:
        return self.get("layer_naming_format", LayerNamingFormat.channel_only)

    @layer_naming_format.setter
    def layer_naming_format(self, value: LayerNamingFormat):
        self.set("layer_naming_format", value)


def get_settings() -> PartSegNapariSettings:
    global _SETTINGS  # noqa: PLW0603  # pylint: disable=global-statement
    if _SETTINGS is None:
        napari_settings = napari_get_settings()
        if hasattr(napari_settings, "path"):
            save_path = napari_settings.path
        else:
            save_path = os.path.dirname(napari_settings.config_path)
        _SETTINGS = PartSegNapariSettings(os.path.join(save_path, "PartSeg_napari_plugins"))
        _SETTINGS.load()
    return _SETTINGS


class SettingsEditor(Container):
    def __init__(self):
        super().__init__()
        self.settings = get_settings()
        self.units_select = create_widget(self.settings.io_units, annotation=Units, label="Units for io")
        self.units_select.changed.connect(self.units_selection_changed)
        self.settings.connect_("io_units", self.units_changed)
        self.append(self.units_select)
        self.layer_naming_select = create_widget(
            self.settings.layer_naming_format, annotation=LayerNamingFormat, label="Format for Layer Name"
        )
        self.layer_naming_select.changed.connect(self.layer_naming_format_selection_changed)
        self.settings.connect_("layer_naming_format", self.layer_naming_format_changed)
        self.append(self.layer_naming_select)

    def units_selection_changed(self, value):
        self.settings.io_units = value
        self.settings.dump()

    def units_changed(self):
        self.units_select.value = self.settings.io_units

    def layer_naming_format_selection_changed(self, value):
        self.settings.layer_naming_format = value
        self.settings.dump()

    def layer_naming_format_changed(self):
        self.layer_naming_select.value = self.settings.layer_naming_format



import numpy as np
from napari import Viewer
from napari.layers.labels import Labels
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import (
    QCheckBox,
    QGridLayout,
    QLabel,
    QPushButton,
    QShortcut,
    QSpinBox,
    QWidget,
)
from superqt.utils import QSignalDebouncer

from PartSeg.common_gui.flow_layout import FlowLayout


class CopyLabelsWidget(QWidget):
    def __init__(self, napari_viewer: Viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.label = QLabel("Copy selected labels along z-axis")

        self.copy_btn = QPushButton("Copy")
        self.check_all_btn = QPushButton("Check all")
        self.uncheck_all_btn = QPushButton("Uncheck all")
        self.lower = QSpinBox()
        self.lower.setSingleStep(1)
        self.upper = QSpinBox()
        self.upper.setSingleStep(1)
        self.checkbox_layout = FlowLayout()
        self._components = set()

        layout = QGridLayout()
        layout.addWidget(self.label, 0, 0, 1, 2)
        layout.setRowStretch(0, 0)
        layout.addWidget(QLabel("Lower layer"), 1, 0)
        layout.addWidget(QLabel("Upper layer"), 2, 0)
        layout.addWidget(self.lower, 1, 1)
        layout.addWidget(self.upper, 2, 1)
        layout.addLayout(self.checkbox_layout, 3, 0, 1, 2)
        layout.setRowStretch(3, 1)
        layout.addWidget(self.copy_btn, 4, 0, 1, 2)
        layout.addWidget(self.check_all_btn, 5, 0)
        layout.addWidget(self.uncheck_all_btn, 5, 1)
        self.shortcut = QShortcut(QKeySequence("Ctrl+K"), self)
        self.shortcut.activated.connect(self.copy_action)
        self.debouncer = QSignalDebouncer(parent=self)
        self.debouncer.setTimeout(200)
        self.debouncer.triggered.connect(self.update_items)

        self.setLayout(layout)

        self.viewer.layers.selection.events.active.connect(self.activate_widget)
        self.copy_btn.clicked.connect(self.copy_action)
        self.check_all_btn.clicked.connect(self._check_all)
        self.uncheck_all_btn.clicked.connect(self._uncheck_all)
        if isinstance(self.viewer.layers.selection.active, Labels):
            self._activate_widget(self.viewer.layers.selection.active)

    def activate_widget(self, event):
        self._activate_widget(event.value)

    def _activate_widget(self, label_layer: Labels):
        is_labels = isinstance(label_layer, Labels)
        if is_labels and not self.isVisible():
            self._update_items(label_layer)
        self.setVisible(is_labels)
        if is_labels:
            label_layer.events.set_data.connect(self._shallow_update)
            label_layer.events.selected_label.connect(self.debouncer.throttle)

    def _shallow_update(self, event):
        label_num = event.source.selected_label
        self._components.add(label_num)
        self.refresh()

    def update_items(self):
        self._update_items(self.viewer.layers.selection.active)

    def _update_items(self, layer):
        unique = np.unique(layer.data)
        if unique[0] == 0:
            unique = unique[1:]
        self._components = set(unique)
        self.refresh()

    def _check_all(self):
        for i in range(self.checkbox_layout.count()):
            w: QCheckBox = self.checkbox_layout.itemAt(i).widget()
            w.setChecked(True)

    def _uncheck_all(self):
        for i in range(self.checkbox_layout.count()):
            w: QCheckBox = self.checkbox_layout.itemAt(i).widget()
            w.setChecked(False)

    def refresh(self):
        checked = set()
        for _ in range(self.checkbox_layout.count()):
            w: QCheckBox = self.checkbox_layout.takeAt(0).widget()
            if w.isChecked():
                checked.add(w.text())
            w.deleteLater()
        for v in self._components:
            chk = QCheckBox(str(v))
            if chk.text() in checked:
                chk.setChecked(True)
            self.checkbox_layout.addWidget(chk)

    def copy_action(self):
        layer = self.viewer.layers.selection.active
        if layer is None:
            return

        leading_zeros = (0,) * (layer.data.ndim - 3)
        checked = set()
        for i in range(self.checkbox_layout.count()):
            w: QCheckBox = self.checkbox_layout.itemAt(i).widget()
            if w.isChecked():
                checked.add(int(w.text()))
        if not checked:
            checked = {layer.selected_label}
        z_position = self.viewer.dims.current_step[layer.dtype.ndim - 3]
        for component_num in checked:
            mask = layer.data[(*leading_zeros, z_position)] == component_num
            start = max(0, self.lower.value())
            end = min(layer.data.shape[1] - 1, self.upper.value()) + 1
            for i in range(start, end):
                layer.data[(*leading_zeros, i)][mask] = component_num



# pylint: disable=no-self-use
from unittest.mock import patch

import numpy as np
import pytest
from qtpy.QtCore import QEvent
from qtpy.QtWidgets import QApplication, QCheckBox

from PartSeg._roi_analysis.measurement_widget import MeasurementsStorage, MeasurementWidget
from PartSeg._roi_mask.simple_measurements import SimpleMeasurements
from PartSegCore.analysis.measurement_base import AreaType, PerComponent
from PartSegCore.analysis.measurement_calculation import ComponentsInfo, MeasurementResult


class TestMeasurementWidget:
    @patch("PartSeg._roi_analysis.measurement_widget.QMessageBox")
    def test_missed_mask(self, qmessagebox_path, qtbot, analysis_segmentation, part_settings, monkeypatch):
        widget = MeasurementWidget(part_settings)
        qtbot.addWidget(widget)

        assert widget.measurement_type.count() == 4
        part_settings.set_project_info(analysis_segmentation)

        with qtbot.waitSignal(widget.measurement_type.currentIndexChanged):
            widget.measurement_type.setCurrentIndex(2)
        assert widget.measurement_type.currentIndex() == 0
        assert not widget.recalculate_button.isEnabled()
        qmessagebox_path.information.assert_called_once()
        assert qmessagebox_path.information.call_args[0][1] == "Need mask"

    @pytest.mark.enablethread
    @pytest.mark.enabledialog
    def test_base(self, qtbot, analysis_segmentation, part_settings):
        widget = MeasurementWidget(part_settings)
        qtbot.addWidget(widget)

        assert widget.measurement_type.count() == 4
        part_settings.set_project_info(analysis_segmentation)
        widget.measurement_type.setCurrentIndex(1)
        assert widget.recalculate_button.isEnabled()
        widget.recalculate_button.click()
        assert widget.info_field.columnCount() == 2
        assert widget.info_field.rowCount() == 2
        assert widget.info_field.item(1, 1).text() == "4"
        widget.horizontal_measurement_present.setChecked(True)
        assert widget.info_field.columnCount() == 2
        assert widget.info_field.rowCount() == 2

    @pytest.mark.enablethread
    @pytest.mark.enabledialog
    def test_base2(self, qtbot, analysis_segmentation2, part_settings):
        widget = MeasurementWidget(part_settings)
        qtbot.addWidget(widget)

        assert widget.measurement_type.count() == 4
        part_settings.set_project_info(analysis_segmentation2)
        widget.measurement_type.setCurrentIndex(2)
        assert widget.recalculate_button.isEnabled()
        widget.recalculate_button.click()
        assert widget.info_field.columnCount() == 2
        assert widget.info_field.rowCount() == 3
        assert widget.info_field.item(1, 1).text() == "4"
        widget.horizontal_measurement_present.setChecked(True)
        assert widget.info_field.columnCount() == 3
        assert widget.info_field.rowCount() == 2

    @pytest.mark.enablethread
    @pytest.mark.enabledialog
    def test_base_channels(self, qtbot, analysis_segmentation2, part_settings):
        widget = MeasurementWidget(part_settings)
        qtbot.addWidget(widget)

        part_settings.set_project_info(analysis_segmentation2)
        widget.measurement_type.setCurrentIndex(3)
        assert widget.recalculate_button.isEnabled()
        widget.recalculate_button.click()
        assert widget.info_field.columnCount() == 2
        assert widget.info_field.rowCount() == 4
        assert widget.info_field.item(1, 1).text() == "4"
        widget.horizontal_measurement_present.setChecked(True)
        assert widget.info_field.columnCount() == 4
        assert widget.info_field.rowCount() == 2


class TestSimpleMeasurementsWidget:
    @pytest.mark.enablethread
    @pytest.mark.enabledialog
    def test_base(self, stack_settings, stack_segmentation1, qtbot):
        widget = SimpleMeasurements(stack_settings)
        qtbot.addWidget(widget)
        stack_settings.set_project_info(stack_segmentation1)
        widget.show()
        event = QEvent(QEvent.WindowActivate)
        QApplication.sendEvent(widget, event)
        assert widget.measurement_layout.count() > 2
        for i in range(2, widget.measurement_layout.count()):
            chk = widget.measurement_layout.itemAt(i).widget()
            assert isinstance(chk, QCheckBox)
            chk.setChecked(True)
        widget.calculate()
        assert widget.result_view.rowCount() == widget.measurement_layout.count() - 1
        assert widget.result_view.columnCount() == len(stack_settings.roi_info.bound_info) + 1


class TestMeasurementsStorage:
    def test_empty(self):
        obj = MeasurementsStorage()
        assert obj.get_size(True) == (0, 0)
        assert obj.get_size(False) == (0, 0)

    def test_base(self):
        info = ComponentsInfo(np.arange(1, 4), np.array([]), {})
        storage = MeasurementResult(info)
        storage["aa"] = 1, "", (PerComponent.No, AreaType.ROI)
        storage["bb"] = [4, 5, 6], "np", (PerComponent.Yes, AreaType.ROI)
        obj = MeasurementsStorage()
        obj.add_measurements(storage)
        obj.set_expand(False)
        assert obj.get_size(True) == (2, 2)
        obj.set_expand(True)
        assert obj.get_size(True) == (3, 4)
        assert obj.get_size(False) == (4, 3)
        obj.set_expand(False)
        assert obj.get_val_as_str(0, 0, True) == "aa"
        assert obj.get_val_as_str(0, 0, False) == "aa"
        assert obj.get_val_as_str(0, 1, True) == "bb"
        assert obj.get_val_as_str(1, 0, False) == "bb"
        assert obj.get_header(True) == ["0", "1"]
        assert obj.get_header(False) == ["Name", "Value"]
        assert obj.get_rows(False) == ["0", "1"]
        assert obj.get_rows(True) == ["Name", "Value"]
        obj.set_expand(True)
        assert obj.get_header(False) == ["Name", "Value", "Value", "Value"]



import dataclasses
import os
import typing
from copy import deepcopy
from io import BytesIO
from pathlib import Path

import numpy as np

from PartSegCore.algorithm_describe_base import AlgorithmProperty
from PartSegCore.analysis.io_utils import ProjectTuple
from PartSegCore.analysis.save_functions import SaveCmap
from PartSegCore.io_utils import SaveBase, SaveROIAsNumpy, SaveROIAsTIFF
from PartSegCore.roi_info import ROIInfo
from PartSegCore.universal_const import Units
from PartSegImage import Channel


class SaveModeling(SaveBase):
    @classmethod
    def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
        return [
            AlgorithmProperty("channel", "Channel", 0, value_type=Channel),
            AlgorithmProperty("clip", "Clip area", False),
            AlgorithmProperty(
                "reverse", "Reverse", False, help_text="Reverse brightness off image (for electron microscopy)"
            ),
            AlgorithmProperty("units", "Units", Units.nm, value_type=Units),
        ]

    @classmethod
    def get_name(cls):
        return "Modeling Data"

    @classmethod
    def get_default_extension(cls):
        return ""

    @classmethod
    def get_short_name(cls):
        return "modeling data"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: ProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        if not os.path.exists(save_location):
            os.makedirs(save_location)
        if not os.path.isdir(save_location):
            raise OSError("save location exist and is not a directory")
        parameters = deepcopy(parameters)
        if parameters["clip"]:
            points = np.nonzero(project_info.roi_info.roi)
            lower_bound = np.min(points, axis=1)
            lower_bound = np.max([lower_bound - 3, [0, 0, 0]], axis=0)
            upper_bound = np.max(points, axis=1)
            upper_bound = np.max([upper_bound + 3, np.array(project_info.roi_info.roi) - 1], axis=0)
            cut_area = tuple(slice(x, y) for x, y in zip(lower_bound, upper_bound))
            # WARNING time
            image = project_info.image.cut_image((slice(None), *cut_area))
            roi_info = ROIInfo(project_info.roi_info.roi[cut_area])

            mask = project_info.mask[cut_area] if project_info.mask else None
            project_info = dataclasses.replace(project_info, image=image, roi_info=roi_info, mask=mask)
            parameters["clip"] = False

        parameters.update({"separated_objects": False})
        SaveCmap.save(
            os.path.join(save_location, "density.cmap"), project_info, parameters, range_changed, step_changed
        )
        parameters.update({"separated_objects": True})
        SaveCmap.save(
            os.path.join(save_location, "density.cmap"), project_info, parameters, range_changed, step_changed
        )
        SaveROIAsTIFF.save(
            os.path.join(save_location, "segmentation.tiff"), project_info, {}, range_changed, step_changed
        )
        SaveROIAsNumpy.save(
            os.path.join(save_location, "segmentation.npy"), project_info, {}, range_changed, step_changed
        )



# pylint: disable=no-self-use
import gc
from functools import partial
from importlib.metadata import version
from unittest.mock import MagicMock

import numpy as np
import pytest
from napari.layers import Image as NapariImage
from napari.qt import QtViewer
from packaging.version import parse as parse_version
from qtpy.QtCore import QPoint
from vispy.geometry import Rect

from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.napari_image_view import (
    ORDER_DICT,
    ImageInfo,
    ImageView,
    QMenu,
    SearchComponentModal,
    SearchType,
    _calc_layer_filter,
    _print_dict,
)
from PartSegCore.image_operations import NoiseFilterType
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image

NAPARI_GE_5_0 = parse_version(version("napari")) >= parse_version("0.5.0a1")
NAPARI_GE_4_19 = parse_version(version("napari")) >= parse_version("0.4.19a1")


if NAPARI_GE_5_0:
    EXPECTED_RANGE = (0, 0, 1)
else:
    EXPECTED_RANGE = (0, 1, 1)

if NAPARI_GE_4_19:

    def get_color_dict(layer):
        return layer.colormap.color_dict

else:

    def get_color_dict(layer):
        return layer.color


def test_image_info():
    image_info = ImageInfo(Image(np.zeros((10, 10)), spacing=(1, 1), axes_order="XY"), [])
    assert not image_info.coords_in([1, 1])
    assert np.all(image_info.translated_coords([1, 1]) == [1, 1])

    image_info.layers.append(NapariImage(image_info.image.get_channel(0), scale=(1, 1, 10, 10)))
    assert image_info.coords_in([0.5, 0.5, 1, 1])
    assert np.all(image_info.translated_coords([1, 1, 1, 1]) == [1, 1, 1, 1])


def test_print_dict():
    dkt = {"a": 1, "b": {"e": 1, "d": [1, 2, 4]}}
    res = _print_dict(dkt)
    lines = res.split("\n")
    assert len(lines) == 4
    assert lines[0].startswith("a")
    assert lines[2].startswith("  e")


@pytest.fixture
def image_view(base_settings, image2, qtbot, request):
    ch_prop = ChannelProperty(base_settings, "test")
    view = ImageView(base_settings, channel_property=ch_prop, name=request.function.__name__)
    base_settings.image = image2
    yield view
    ch_prop.close()
    view.close()
    ch_prop.deleteLater()
    view.deleteLater()
    qtbot.wait(50)
    gc.collect()
    if hasattr(QtViewer, "_instances"):
        QtViewer._instances.clear()


class TestImageView:
    def test_constructor(self, base_settings, qtbot):
        ch_prop = ChannelProperty(base_settings, "test")
        view = ImageView(base_settings, channel_property=ch_prop, name="test")
        qtbot.addWidget(ch_prop)
        qtbot.addWidget(view)

    def test_print_info(self, base_settings, image2, qtbot, monkeypatch):
        ch_prop = ChannelProperty(base_settings, "test")
        view = ImageView(base_settings, channel_property=ch_prop, name="test")
        qtbot.addWidget(ch_prop)
        qtbot.addWidget(view)
        with qtbot.assert_not_emitted(view.text_info_change):
            view.print_info()
        base_settings.image = image2
        monkeypatch.setattr(view, "_coordinates", lambda: (-10, 0.5, 0.5, 0.5))
        with qtbot.waitSignal(view.text_info_change, check_params_cb=lambda x: x == ""):
            view.print_info()
        base_settings.roi = np.ones(image2.get_channel(0).shape, dtype=np.uint8)
        monkeypatch.setattr(view, "_coordinates", lambda: (0.5, 0.5, 0.5, 0.5))
        with qtbot.waitSignal(view.text_info_change, check_params_cb=lambda x: "component: " in x):
            view.print_info()

    def test_synchronize_viewers(self, base_settings, image2, qtbot):
        ch_prop = ChannelProperty(base_settings, "test")
        view1 = ImageView(base_settings, channel_property=ch_prop, name="test")
        view2 = ImageView(base_settings, channel_property=ch_prop, name="test2")
        qtbot.addWidget(ch_prop)
        qtbot.addWidget(view1)
        qtbot.addWidget(view2)
        base_settings.image = image2

        view1.viewer.dims.set_point(1, 5)
        view2.set_state(view1.get_state())
        view1.viewer.dims.ndisplay = 3
        view2.set_state(view1.get_state())

    def test_order_reset(self, image_view):
        image_view._rotate_dim()
        assert image_view.viewer.dims.order == (0, 2, 1, 3)
        image_view._reset_view()
        assert image_view.viewer.dims.order == (0, 1, 2, 3)

    def test_update_rendering(self, base_settings, image_view):
        base_settings.roi = np.ones(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        image_view.update_rendering()

    def test_roi_rendering(self, base_settings, image_view, tmp_path):
        roi = np.ones(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 1, 1] = 0
        image_view.set_roi(ROIInfo(None))
        base_settings.roi = roi
        image_view.update_roi_border()
        assert np.any(image_view.image_info[str(tmp_path / "test2.tiff")].roi.data == 1)
        image_view.settings.set_in_profile(f"{image_view.name}.image_state.only_border", False)
        image_view.update_roi_border()
        assert np.count_nonzero(image_view.image_info[str(tmp_path / "test2.tiff")].roi.data) == 20**3 - 20
        base_settings.roi = np.ones(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].roi.data)
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].roi.scale == (1, 10**6, 10**6, 10**6))
        base_settings.image.set_spacing((10**-4,) * 3)
        image_view.update_spacing_info()
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].roi.scale == (1, 10**5, 10**5, 10**5))
        base_settings.roi = None
        assert not image_view.image_info[str(tmp_path / "test2.tiff")].roi.visible
        base_settings.roi = roi
        assert image_view.image_info[str(tmp_path / "test2.tiff")].roi.visible
        image_view.remove_all_roi()
        assert not image_view.image_info[str(tmp_path / "test2.tiff")].roi.visible

    def test_has_image(self, base_settings, image_view, image, image2):
        base_settings.image = image2
        assert image_view.has_image(image2)
        assert not image_view.has_image(image)
        image_view.update_spacing_info()

    def test_mask_rendering(self, base_settings, image_view, qtbot, tmp_path):
        image_view.set_mask()
        with qtbot.waitSignal(base_settings.mask_changed):
            base_settings.mask = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].mask.data == 1)
        # double add for test if proper refresh mask is working
        with qtbot.waitSignal(base_settings.mask_changed):
            base_settings.mask = np.ones(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].mask.data == 0)
        base_settings.set_in_profile("mask_presentation_opacity", 0.5)
        assert image_view.image_info[str(tmp_path / "test2.tiff")].mask.opacity == 0.5
        base_settings.set_in_profile("mask_presentation_color", (255, 0, 0))
        assert np.all(get_color_dict(image_view.image_info[str(tmp_path / "test2.tiff")].mask)[1] == (1, 0, 0, 1))
        base_settings.set_in_profile("mask_presentation_color", (128, 0, 0))
        assert np.allclose(
            get_color_dict(image_view.image_info[str(tmp_path / "test2.tiff")].mask)[1], (128 / 255, 0, 0, 1)
        )

        assert not image_view.image_info[str(tmp_path / "test2.tiff")].mask.visible
        with qtbot.waitSignal(image_view.mask_chk.stateChanged):
            image_view.mask_chk.setChecked(True)
        assert image_view.image_info[str(tmp_path / "test2.tiff")].mask.visible
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].mask.scale == (1, 10**6, 10**6, 10**6))
        base_settings.image.set_spacing((10**-4,) * 3)
        image_view.update_spacing_info()
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].mask.scale == (1, 10**5, 10**5, 10**5))

    @pytest.mark.windows_ci_skip
    def test_mask_control_visibility(self, base_settings, image_view, qtbot):
        image_view.show()
        assert not image_view.mask_chk.isVisible()
        image_view.set_mask()
        assert not image_view.mask_chk.isVisible()
        with qtbot.waitSignal(base_settings.mask_changed):
            base_settings.mask = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        assert image_view.mask_chk.isVisible()
        with qtbot.waitSignal(base_settings.mask_changed):
            base_settings.mask = None
        assert not image_view.mask_chk.isVisible()
        image_view.hide()

    def test_points_rendering(self, base_settings, image_view):
        assert image_view.points_layer is None
        base_settings.points = [(0, 5, 5, 5)]
        assert image_view.points_layer is not None
        assert image_view.points_layer.visible
        image_view.toggle_points_visibility()
        assert not image_view.points_layer.visible

    @pytest.mark.windows_ci_skip
    def test_points_button_visibility(self, base_settings, image_view):
        image_view.show()
        assert not image_view.points_view_button.isVisible()
        base_settings.points = [(0, 5, 5, 5)]
        assert image_view.points_view_button.isVisible()
        base_settings.points = None
        assert not image_view.points_view_button.isVisible()
        image_view.hide()

    def test_dim_menu(self, image_view, monkeypatch):
        called = []

        from PartSeg.common_gui import napari_image_view

        def check_menu(self, point):
            assert len(self.actions()) == len(ORDER_DICT)
            called.append(1)

        def _monkey_qmenu():
            res = QMenu()
            monkeypatch.setattr(res, "exec_", partial(check_menu, res))
            return res

        monkeypatch.setattr(napari_image_view, "QMenu", _monkey_qmenu)
        image_view._dim_order_menu(QPoint(0, 0))
        assert called == [1]

    def test_update_alternatives(self, base_settings, image_view, tmp_path):
        roi = np.ones(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 1, 1] = 0
        base_settings.roi = ROIInfo(roi, alternative={"test": np.ones(roi.shape, dtype=np.uint8) - roi})
        image_view.update_roi_border()
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].roi.data == roi)
        image_view.roi_alternative_selection = "test"
        image_view.update_roi_border()
        assert np.all(image_view.image_info[str(tmp_path / "test2.tiff")].roi.data != roi)

    def test_marking_component(self, base_settings, image_view, tmp_path):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        image_view.component_mark(1, False)
        assert image_view.image_info[str(tmp_path / "test2.tiff")].highlight is not None
        assert image_view.image_info[str(tmp_path / "test2.tiff")].highlight.visible
        assert image_view.image_info[str(tmp_path / "test2.tiff")].highlight in image_view.viewer.layers
        assert "timer" not in image_view.image_info[str(tmp_path / "test2.tiff")].highlight.metadata
        image_view.component_unmark(0)
        assert not image_view.image_info[str(tmp_path / "test2.tiff")].highlight.visible
        image_view.component_mark(10, False)
        assert not image_view.image_info[str(tmp_path / "test2.tiff")].highlight.visible

    @pytest.mark.enablethread
    def test_marking_component_flash(self, base_settings, image_view, tmp_path, qtbot):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        image_view.component_mark(1, True)
        assert image_view.image_info[str(tmp_path / "test2.tiff")].highlight.visible
        assert "timer" in image_view.image_info[str(tmp_path / "test2.tiff")].highlight.metadata
        timer = image_view.image_info[str(tmp_path / "test2.tiff")].highlight.metadata["timer"]
        assert timer.isActive()
        assert image_view.viewer.dims.range[0] == EXPECTED_RANGE
        qtbot.wait(800)
        image_view.component_unmark(0)
        assert not image_view.image_info[str(tmp_path / "test2.tiff")].highlight.visible
        assert not timer.isActive()

    @pytest.mark.parametrize("pos", [(-(10**8), -(10**8)), (10**8, 10**8)])
    def test_update_camera_position_component_marking(self, base_settings, image_view, pos):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        image_view.viewer.dims.set_point(1, 0)
        rect = Rect(image_view.viewer_widget.view.camera.get_state()["rect"])
        rect.pos = pos
        rect.size = (100, 100)
        image_view.viewer_widget.view.camera.set_state({"rect": rect})
        image_view.component_mark(1, False)
        assert image_view.viewer.dims.point[1] != 0
        assert image_view.viewer_widget.view.camera.get_state()["rect"].pos != pos

    def test_zoom_mark(self, base_settings, image_view):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        image_view.component_zoom(1)
        image_view.component_zoom(10)

    def test_roi_removed_restore_roi(self, base_settings, image_view):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        assert "ROI" in image_view.viewer.layers
        del image_view.viewer.layers["ROI"]
        assert "ROI" not in image_view.viewer.layers
        image_view.set_roi()
        assert "ROI" in image_view.viewer.layers

    def test_roi_removed_add_image_restore(self, base_settings, image_view):
        roi = np.zeros(base_settings.image.get_channel(0).shape, dtype=np.uint8)
        roi[..., 2:-2, 2:-2, 2:-2] = 1
        base_settings.roi = roi
        base_settings.mask = roi
        assert "ROI" in image_view.viewer.layers
        assert "Mask" in image_view.viewer.layers
        image_view._add_image((image_view.image_info[base_settings.image.file_path], True))
        assert "ROI" in image_view.viewer.layers
        assert "Mask" in image_view.viewer.layers

    @pytest.mark.parametrize("filter_type", NoiseFilterType.__members__.values())
    @pytest.mark.parametrize("radius", [1, 2, 3.5])
    def test_calculate_filter(self, filter_type, radius, image):
        ch = image.get_channel(0)
        filtered = ImageView.calculate_filter(ch, (filter_type, radius))
        assert filtered.shape == ch.shape
        assert (filter_type == NoiseFilterType.No) != (filtered is not ch)

    @pytest.mark.no_patch_add_layer
    @pytest.mark.enablethread
    def test_add_layer_util_check_init(self, image_view, qtbot):
        def has_layers():
            return len(image_view.viewer.layers) > 0

        qtbot.waitUntil(has_layers)

    @pytest.mark.no_patch_add_layer
    def test_add_layer_util(self, image_view, qtbot):
        def has_layers():
            return len(image_view.viewer.layers) > 0

        qtbot.waitUntil(has_layers)

        layer = image_view.viewer.layers[0]
        del image_view.viewer.layers[layer.name]
        assert isinstance(layer, NapariImage)
        layer.visible = False

        def no_worker():
            return len(image_view.worker_list) == 0

        prev_data = layer.data

        image_view._add_layer_util(0, layer, [(NoiseFilterType.No, 0)])

        qtbot.waitUntil(no_worker)
        assert layer.data is prev_data

        del image_view.viewer.layers[layer.name]

        image_view._add_layer_util(0, layer, [(NoiseFilterType.Median, 1)])

        qtbot.waitUntil(no_worker)
        assert layer.data is not prev_data
        qtbot.wait(50)


def test_search_component_modal(qtbot, image_view, monkeypatch):
    monkeypatch.setattr(image_view, "component_mark", MagicMock())
    monkeypatch.setattr(image_view, "component_zoom", MagicMock())
    monkeypatch.setattr(image_view, "component_unmark", MagicMock())
    modal = SearchComponentModal(image_view, SearchType.Highlight, component_num=1, max_components=5)
    qtbot.addWidget(modal)
    image_view.component_mark.assert_called_with(1, flash=True)
    modal.component_selector.setValue(2)
    image_view.component_mark.assert_called_with(2, flash=True)
    assert image_view.component_zoom.call_count == 0
    modal.zoom_to.setCurrentEnum(SearchType.Zoom_in)
    image_view.component_zoom.assert_called_with(2)
    assert image_view.component_mark.call_count == 2
    modal.component_selector.setValue(1)
    image_view.component_zoom.assert_called_with(1)
    modal.close()
    image_view.component_unmark.assert_called_once()


def test_calc_layer_filter():
    layer = NapariImage(np.zeros((10, 10)), name="test")

    assert _calc_layer_filter(layer, NoiseFilterType.No, 0)[0] is None
    assert _calc_layer_filter(layer, NoiseFilterType.No, 1)[0] is None
    assert _calc_layer_filter(layer, NoiseFilterType.Gauss, 0)[0] is None

    assert _calc_layer_filter(layer, NoiseFilterType.Gauss, 1)[0] is not None
    assert _calc_layer_filter(layer, NoiseFilterType.Gauss, 1)[0] is not layer.data



from io import BytesIO
from unittest.mock import MagicMock, Mock

import pytest

from PartSeg._launcher.check_survey import CheckSurveyThread, SurveyMessageBox


@pytest.fixture(autouse=True)
def _mock_file_path(monkeypatch, tmp_path):
    monkeypatch.setattr("PartSeg._launcher.check_survey.IGNORE_FILE_PATH", tmp_path / "file1.txt")


@pytest.mark.usefixtures("qtbot")
def test_check_survey_thread(monkeypatch, tmp_path):
    urlopen_mock = Mock(return_value=BytesIO(b"some data"))
    monkeypatch.setattr("urllib.request.urlopen", urlopen_mock)
    thr = CheckSurveyThread("test_url")

    thr.run()
    assert urlopen_mock.call_count == 1
    assert thr.survey_url == "some data"

    (tmp_path / "file1.txt").touch()
    thr2 = CheckSurveyThread("test_url")
    thr2.run()

    assert urlopen_mock.call_count == 1
    assert thr2.survey_url == ""


@pytest.mark.usefixtures("qtbot")
def test_thread_ignore_file_exist(monkeypatch, tmp_path):
    message_mock = Mock()
    monkeypatch.setattr("PartSeg._launcher.check_survey.SurveyMessageBox", message_mock)
    with (tmp_path / "file1.txt").open("w") as f_p:
        f_p.write("test_url")

    thr = CheckSurveyThread("test_url2")
    thr.survey_url = "test_url"

    thr.show_version_info()
    assert message_mock.call_count == 0


@pytest.mark.usefixtures("qtbot")
def test_call_survey_message_box(monkeypatch):
    message_mock = MagicMock()
    monkeypatch.setattr("PartSeg._launcher.check_survey.SurveyMessageBox", message_mock)
    thr = CheckSurveyThread("test_url")
    thr.survey_url = "test_url"
    thr.show_version_info()
    assert message_mock.call_count == 1


def test_survey_message_box(qtbot, monkeypatch, tmp_path):
    web_open = Mock()
    monkeypatch.setattr("PartSeg._launcher.check_survey.webbrowser.open", web_open)
    monkeypatch.setattr("PartSeg._launcher.check_survey.QMessageBox.exec_", Mock(return_value=0))
    msg = SurveyMessageBox("test_url")
    qtbot.addWidget(msg)

    msg._open_btn.click()
    web_open.assert_called_once_with("test_url")

    msg.exec_()
    assert (tmp_path / "file1.txt").exists()
    assert (tmp_path / "file1.txt").read_text() == ""

    msg._ignore_btn.click()

    assert (tmp_path / "file1.txt").exists()
    assert (tmp_path / "file1.txt").read_text() == "test_url"



import warnings

import napari
import numpy as np
from magicgui.widgets import Container, HBox, PushButton, SpinBox, create_widget
from napari import Viewer
from napari.layers import Labels
from napari.utils.notifications import show_info
from packaging.version import parse as parse_version
from qtpy.QtCore import QTimer
from vispy.geometry import Rect

from PartSeg.common_gui.napari_image_view import SearchType, get_highlight_colormap
from PartSegCore.roi_info import ROIInfo

HIGHLIGHT_LABEL_NAME = ".Highlight"

_napari_ge_5 = parse_version(napari.__version__) >= parse_version("0.5.0a1")


class SearchLabel(Container):
    def __init__(self, napari_viewer: Viewer):
        super().__init__(layout="vertical")
        self.napari_viewer = napari_viewer
        self.search_type = create_widget(annotation=SearchType, label="Search type")
        self.search_type.changed.connect(self._component_num_changed)
        self.component_selector = SpinBox(name="Label number", min=0, max=2**31 - 1)
        self.component_selector.changed.connect(self._component_num_changed)
        self.labels_layer = create_widget(annotation=Labels, label="ROI", options={})
        self.labels_layer.changed.connect(self._update_roi_info)
        self.zoom_factor = create_widget(annotation=float, label="Zoom factor", value=1.2)
        self.zoom_factor.changed.connect(self._component_num_changed)
        self.stop = PushButton(name="Stop")
        self.stop.clicked.connect(self._stop)
        self.roi_info = None

        layout = HBox(
            widgets=(
                self.search_type,
                self.component_selector,
            )
        )
        layout2 = HBox(
            widgets=(
                self.labels_layer,
                self.zoom_factor,
            )
        )
        self.insert(0, layout)
        self.insert(1, layout2)
        self.insert(2, self.stop)

        self._update_roi_info()

    def _update_roi_info(self):
        if self.labels_layer.value is None:
            return
        self.roi_info = ROIInfo(self.labels_layer.value.data)
        self._component_num_changed()

    def _stop(self):
        if HIGHLIGHT_LABEL_NAME in self.napari_viewer.layers:
            self.napari_viewer.layers[HIGHLIGHT_LABEL_NAME].metadata["timer"].stop()
            del self.napari_viewer.layers[HIGHLIGHT_LABEL_NAME]

    def _highlight(self):
        num = self.component_selector.value
        labels = self.labels_layer.value
        bound_info = self.roi_info.bound_info.get(num, None)
        if bound_info is None:
            self._stop()
            return
        slices = bound_info.get_slices()
        component_mark = self.roi_info.roi[tuple(slices)] == num
        translate_grid = labels.translate + bound_info.lower * labels.scale
        if HIGHLIGHT_LABEL_NAME in self.napari_viewer.layers:
            self.napari_viewer.layers[HIGHLIGHT_LABEL_NAME].data = component_mark
        else:
            layer = self.napari_viewer.add_labels(
                component_mark,
                name=HIGHLIGHT_LABEL_NAME,
                scale=labels.scale,
                blending="translucent",
                opacity=0.7,
                **get_highlight_colormap(),
            )

            def flash_fun(layer_=layer):
                opacity = layer_.opacity + 0.1
                if opacity > 1:
                    opacity = 0.1
                layer_.opacity = opacity

            timer = QTimer()
            timer.setInterval(100)
            timer.timeout.connect(flash_fun)
            timer.start()
            layer.metadata["timer"] = timer

        self.napari_viewer.layers[HIGHLIGHT_LABEL_NAME].translate = translate_grid
        self._shift_if_need(labels, bound_info)

    def _shift_if_need(self, labels, bound_info):
        if self.napari_viewer.dims.ndisplay != 2:
            return
        lower_bound = self._data_to_world(labels, bound_info.lower)
        upper_bound = self._data_to_world(labels, bound_info.upper)
        self._update_point(lower_bound, upper_bound)
        l_bound = lower_bound[-2:][::-1]
        u_bound = upper_bound[-2:][::-1]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Public access to Window.qt_viewer")
            if _napari_ge_5:
                rect = Rect(self.napari_viewer.window.qt_viewer.canvas.view.camera.get_state()["rect"])
            else:
                rect = Rect(self.napari_viewer.window.qt_viewer.view.camera.get_state()["rect"])
        if rect.contains(*l_bound) and rect.contains(*u_bound):
            return
        size = u_bound - l_bound
        rect.size = tuple(np.max([rect.size, size * 1.2], axis=0))
        pos = rect.pos
        if rect.left > l_bound[0]:
            pos = l_bound[0], pos[1]
        if rect.right < u_bound[0]:
            pos = pos[0] + u_bound[0] - rect.right, pos[1]
        if rect.bottom > l_bound[1]:
            pos = pos[0], l_bound[1]
        if rect.top < u_bound[1]:
            pos = pos[0], pos[1] + (u_bound[1] - rect.top)
        rect.pos = pos
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Public access to Window.qt_viewer")
            if _napari_ge_5:
                self.napari_viewer.window.qt_viewer.canvas.view.camera.set_state({"rect": rect})
            else:
                self.napari_viewer.window.qt_viewer.view.camera.set_state({"rect": rect})

    @staticmethod
    def _data_to_world(layer: Labels, cords):
        return layer._transforms[1:3].simplified(cords)  # pylint: disable=protected-access

    def _zoom(self):
        self._stop()
        if self.napari_viewer.dims.ndisplay != 2:
            show_info("Zoom in does not work in 3D mode")
        num = self.component_selector.value
        labels = self.labels_layer.value
        bound_info = self.roi_info.bound_info.get(num, None)
        if bound_info is None:
            return
        lower_bound = self._data_to_world(labels, bound_info.lower)
        upper_bound = self._data_to_world(labels, bound_info.upper)
        diff = upper_bound - lower_bound
        frame = diff * (self.zoom_factor.value - 1)
        if self.napari_viewer.dims.ndisplay == 2:
            rect = Rect(pos=(lower_bound - frame)[-2:][::-1], size=(diff + 2 * frame)[-2:][::-1])
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Public access to Window.qt_viewer")
                if _napari_ge_5:
                    self.napari_viewer.window.qt_viewer.canvas.view.camera.set_state({"rect": rect})
                else:
                    self.napari_viewer.window.qt_viewer.view.camera.set_state({"rect": rect})
        self._update_point(lower_bound, upper_bound)

    def _update_point(self, lower_bound, upper_bound):
        point = (lower_bound + upper_bound) / 2
        current_point = self.napari_viewer.dims.point[-len(lower_bound) :]
        dims = len(lower_bound) - self.napari_viewer.dims.ndisplay
        start_dims = self.napari_viewer.dims.ndim - len(lower_bound)
        for i in range(dims):
            if not (lower_bound[i] <= current_point[i] <= upper_bound[i]):
                self.napari_viewer.dims.set_point(start_dims + i, point[i])

    def _component_num_changed(self):
        if self.roi_info is None:
            return
        if self.search_type.value == SearchType.Highlight:
            self._highlight()
        else:
            self._zoom()



import contextlib
import importlib
import logging
import os
import pkgutil
import sys
import typing
from pathlib import Path

import PartSegCore.plugins


def register_napari_plugins():  # pragma: no cover
    import npe2

    import PartSeg

    with contextlib.suppress(ValueError):
        npe2.PluginManager.instance().register(
            os.path.join(os.path.dirname(os.path.dirname(PartSeg.__file__)), "napari.yaml")
        )
        base_path = Path(__file__).parent.parent.parent / "plugins"
        for el in base_path.glob("*/napari.yaml"):
            try:
                npe2.PluginManager.instance().register(str(el))
            except Exception:  # noqa: PERF203
                logging.exception("Failed to register %s", el)


def get_plugins():
    if getattr(sys, "frozen", False):  # pragma: no cover
        new_path = [os.path.join(os.path.dirname(os.path.dirname(__path__[0])), "plugins")]
        sys.path.append(new_path[0])
        packages = pkgutil.iter_modules(new_path)
        register_napari_plugins()
    else:
        sys.path.append(os.path.dirname(__file__))
        packages = list(pkgutil.iter_modules(__path__))
    packages2 = PartSegCore.plugins.iter_entrypoints("PartSeg.plugins")
    return [importlib.import_module(el.name) for el in packages] + [el.load() for el in packages2]


plugins_loaded = set()


def register():
    PartSegCore.plugins.register()
    for el in get_plugins():
        if hasattr(el, "register") and el.__name__ not in plugins_loaded:
            if not isinstance(el.register, typing.Callable):  # pragma: no cover
                raise TypeError(f"Plugin {el.__name__} has no register method")
            el.register()
            plugins_loaded.add(el.__name__)


def register_if_need():
    if len(plugins_loaded) == 0:
        register()



import copy
import inspect
import itertools
import typing
import warnings
import weakref
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import suppress
from types import MethodType

import numpy as np
from local_migrator import register_class
from psygnal import Signal
from pydantic import BaseModel as PydanticBaseModel
from sentry_sdk.utils import safe_repr as _safe_repr

__author__ = "Grzegorz Bokota"

if typing.TYPE_CHECKING:
    from napari.layers import Image


def bisect(arr, val, comp):
    left = -1
    right = len(arr)
    while right - left > 1:
        mid = (left + right) >> 1
        if comp(arr[mid], val):
            left = mid
        else:
            right = mid
    return right


def numpy_repr(val: np.ndarray):
    if val is None:  # pragma: no cover
        return repr(val)
    if val.size < 20:
        return repr(val)
    return f"array(size={val.size}, shape={val.shape}, dtype={val.dtype}, min={val.min()}, max={val.max()})"


class CallbackBase(ABC):
    @abstractmethod
    def is_alive(self):
        raise NotImplementedError

    @abstractmethod
    def __call__(self, *args, **kwarg):
        raise NotImplementedError


class CallbackFun(CallbackBase):
    def __init__(self, fun: typing.Callable, max_args: typing.Optional[int] = None):
        self.fun = fun
        self.count = _inspect_signature(fun) if max_args is None else max_args

    def is_alive(self):
        return True

    def __call__(self, *args, **kwarg):
        self.fun(*args[: self.count], **kwarg)


class CallbackMethod(CallbackBase):
    def __init__(self, method, max_args: typing.Optional[int] = None):
        obj, name = self._get_proper_name(method)
        self.ref = weakref.ref(obj)
        self.name = name
        self.count = _inspect_signature(method) if max_args is None else max_args

    @staticmethod
    def _get_proper_name(callback):
        obj = callback.__self__
        if not hasattr(obj, callback.__name__) or getattr(obj, callback.__name__) != callback:
            # some decorators will alter method.__name__, so that obj.method
            # will not be equal to getattr(obj, obj.method.__name__). We check
            # for that case here and traverse to find the right method here.
            for name in dir(obj):
                meth = getattr(obj, name)
                if inspect.ismethod(meth) and meth == callback:
                    return obj, name
            raise RuntimeError(f"During bind method {callback} of object {obj} an error happen")
        return obj, callback.__name__

    def is_alive(self):
        return self.ref() is not None

    def __call__(self, *args, **kwarg):
        obj = self.ref()
        if obj is not None:
            getattr(obj, self.name)(*args[: self.count], **kwarg)


def _inspect_signature(slot: typing.Callable) -> typing.Optional[int]:
    """
    count maximal number of positional argument
    :param slot: callable to be checked
    :return: number of parameters which could be passed to callable, None if unbound
    """
    if hasattr(slot, "__module__") and isinstance(slot.__module__, str) and slot.__module__.startswith("superqt"):
        return 0
    try:
        signature = inspect.signature(slot)
    except ValueError:
        return 0
    count = 0
    for parameter in signature.parameters.values():
        if parameter.kind in [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]:
            count += 1
        if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            count = None
            break
    return count


def get_callback(callback: typing.Union[typing.Callable, MethodType], max_args=None) -> CallbackBase:
    if inspect.ismethod(callback):
        return CallbackMethod(callback, max_args)

    return CallbackFun(callback, max_args)


@register_class(old_paths=["PartSegCore.json_hooks.EventedDict"], allow_errors_in_values=True)
class EventedDict(typing.MutableMapping):
    """
    Class for storing data in dictionary with possibility to connect to change of data.

    :param klass: class of stored data. It could be Type or dict with key as name of key and value as class of data.
        Key "*" is used as default class, for key different form specified one. .
    """

    setted = Signal(str)
    deleted = Signal(str)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        cls = self.__class__
        result = cls(self._klass, **copy.deepcopy(self._dict))
        result.base_key = self.base_key
        memodict[id(self)] = result
        return result

    def __init__(self, klass=None, /, **kwargs):
        super().__init__()
        self._dict = {}
        if klass is None:
            klass = object
        self._klass = klass if isinstance(klass, dict) else {"*": klass}

        for key, dkt in kwargs.items():
            self[key] = dkt
        self.base_key = ""

    def __setitem__(self, k, v) -> None:
        klass = self._klass.get(k, self._klass.get("*", object))
        if isinstance(klass, dict):
            if not isinstance(v, typing.MutableMapping):
                raise TypeError(f"Value for key {k} should be dict")
        elif not isinstance(v, klass):
            raise TypeError(f"Value {v} for key {k} is not instance of {klass}")

        if isinstance(v, dict):
            v = EventedDict(klass, **v)
        if isinstance(v, EventedDict):
            v.base_key = k
            v._klass = klass if isinstance(klass, dict) else {"*": klass}
            v.setted.connect(self._propagate_setitem)
            v.deleted.connect(self._propagate_del)
        if k in self._dict and isinstance(self._dict[k], EventedDict):
            self._dict[k].setted.disconnect(self._propagate_setitem)
            self._dict[k].deleted.disconnect(self._propagate_del)
        old_value = self._dict.get(k)
        with suppress(ValueError):
            if old_value == v:
                return
        self._dict[k] = v
        self.setted.emit(k)

    def __delitem__(self, k) -> None:
        if k in self._dict and isinstance(self._dict[k], EventedDict):
            self._dict[k].setted.disconnect(self._propagate_setitem)
            self._dict[k].deleted.disconnect(self._propagate_del)
        del self._dict[k]
        self.deleted.emit(k)

    def __getitem__(self, k):
        return self._dict[k]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> typing.Iterator:
        return iter(self._dict)

    def as_dict(self):
        return copy.copy(self._dict)

    def as_dict_deep(self):
        return {k: v.as_dict_deep() if isinstance(v, EventedDict) else v for k, v in self._dict.items()}

    def __str__(self):
        return f"EventedDict[{self._klass}]({self._dict})"

    def __repr__(self):
        return f"EventedDict(klass={self._klass}, {self._dict!r})"

    def _propagate_setitem(self, key):
        # Fixme when partial disconnect will work
        sender: EventedDict = Signal.sender()
        if sender.base_key:
            self.setted.emit(f"{sender.base_key}.{key}")
        else:
            self.setted.emit(key)

    def _propagate_del(self, key):
        # Fixme when partial disconnect will work
        sender: EventedDict = Signal.sender()
        if sender.base_key:
            self.deleted.emit(f"{sender.base_key}.{key}")
        else:
            self.deleted.emit(key)


def recursive_update_dict(main_dict: typing.MutableMapping, other_dict: typing.MutableMapping):
    """
    recursive update main_dict with elements of other_dict

    :param main_dict: dict to be updated recursively
    :param other_dict: source dict

    >>> dkt1 = {"test": {"test": 1, "test2": {"test4": 1}}}
    >>> dkt2 = {"test": {"test": 4, "test2": {"test2": 1}}}
    >>> recursive_update_dict(dkt1, dkt2)
    >>> dkt1
        {"test": {"test": 1, "test2": {"test4": 1, "test2": 1}}}
    """
    for key, val in other_dict.items():
        if (
            key in main_dict
            and isinstance(main_dict[key], typing.MutableMapping)
            and isinstance(val, typing.MutableMapping)
        ):
            recursive_update_dict(main_dict[key], val)
        else:
            main_dict[key] = val


@register_class(old_paths=["PartSegCore.json_hooks.ProfileDict"], allow_errors_in_values=True)
class ProfileDict:
    """
    Dict for storing recursive data. The path is dot separated.

    :param klass: class of stored data. Same as in :py:class:`EventedDict`
    :param kwargs: initial data

    >>> dkt = ProfileDict()
    >>> dkt.set(["aa", "bb", "c1"], 7)
    >>> dkt.get(["aa", "bb", "c1"])
        7
    >>> dkt.get("aa.bb.c2", 8)
        8
    >>> dkt.get("aa.bb")
        {'c1': 7, 'c2': 8}
    """

    def __init__(self, klass=None, **kwargs):
        self._my_dict = EventedDict(klass, **kwargs)
        self._callback_dict: dict[str, list[CallbackBase]] = defaultdict(list)

        self._my_dict.setted.connect(self._call_callback)
        self._my_dict.deleted.connect(self._call_callback)

    def as_dict(self):
        return self.my_dict.as_dict()

    @property
    def my_dict(self) -> EventedDict:
        return self._my_dict

    @my_dict.setter
    def my_dict(self, value: typing.Union[dict, EventedDict]):
        if isinstance(value, dict):
            value = EventedDict(**value)

        self._my_dict.setted.disconnect(self._call_callback)
        self._my_dict.deleted.disconnect(self._call_callback)
        self._my_dict = value
        self._my_dict.setted.connect(self._call_callback)
        self._my_dict.deleted.connect(self._call_callback)

    def update(self, ob: typing.Union["ProfileDict", dict, None] = None, **kwargs):
        """
        Update dict recursively. Use :py:func:`~.recursive_update_dict`

        :param ob: data source
        :param kwargs: data source, as keywords
        """
        if isinstance(ob, ProfileDict):
            recursive_update_dict(self.my_dict, ob.my_dict)
            recursive_update_dict(self.my_dict, kwargs)
        elif isinstance(ob, dict):
            recursive_update_dict(self.my_dict, ob)
            recursive_update_dict(self.my_dict, kwargs)
        elif ob is None:
            recursive_update_dict(self.my_dict, kwargs)

    def profile_change(self):
        for callback in itertools.chain(*self._callback_dict.values()):
            callback()

    def connect(
        self, key_path: typing.Union[typing.Sequence[str], str], callback: typing.Callable[[], typing.Any], maxargs=None
    ) -> typing.Callable:
        """
        Connect function to receive information when object on path was changed using :py:meth:`.set`

        :param key_path: path for which signal should be emitted
        :param callback: parameterless function which should be called

        :return: callback function itself.
        """
        if not isinstance(key_path, str):
            key_path = ".".join(key_path)

        self._callback_dict[key_path].append(get_callback(callback, maxargs))
        return callback

    def set(self, key_path: typing.Union[typing.Sequence[str], str], value):
        """
        Set value from dict

        :param key_path: Path to element. If is string then will be split on '.' (`key_path.split('.')`)
        :param value: Value to set.
        """
        if isinstance(key_path, str):
            key_path = key_path.split(".")
        curr_dict = self.my_dict
        i = 0
        try:
            for i, key in enumerate(key_path[:-1]):  # noqa: B007
                # TODO add check if next step element is dict and create custom information
                curr_dict = curr_dict[key]
        except KeyError:
            for key2 in key_path[i:-1]:
                with curr_dict.setted.blocked():
                    curr_dict[key2] = {}
                curr_dict = curr_dict[key2]
        if isinstance(value, dict):
            value = EventedDict(**value)
        curr_dict[key_path[-1]] = value
        return curr_dict[key_path[-1]]

    def _call_callback(self, key_path: typing.Union[typing.Sequence[str], str]):
        if isinstance(key_path, str):
            key_path = key_path.split(".")
        full_path = ".".join(key_path[1:])
        callback_path = ""
        callback_list = []
        if callback_path in self._callback_dict:
            callback_list = self._callback_dict[callback_path]

        for callback_path in itertools.accumulate(key_path[1:], lambda x, y: f"{x}.{y}"):
            if callback_path in self._callback_dict:
                li = self._callback_dict[callback_path]
                li = [x for x in li if x.is_alive()]
                self._callback_dict[callback_path] = li
                callback_list.extend(li)
        for callback in callback_list:
            callback(full_path)

    def get(self, key_path: typing.Union[list, str], default=None):
        """
        Get value from dict.

        :param key_path: Path to element. If is string then will be split on . (`key_path.split('.')`).
        :param default: default value if element missed in dict.
        :raise KeyError: on missed element if default is not provided.
        :return: requested value
        """
        if isinstance(key_path, str):
            key_path = key_path.split(".")
        curr_dict = self.my_dict
        try:
            for key in key_path:
                curr_dict = curr_dict[key]
        except KeyError as e:
            if default is None:
                raise e

            val = copy.deepcopy(default)
            return self.set(key_path, val)

        return curr_dict

    def verify_data(self) -> bool:
        """
        Call :py:func:`~.check_loaded_dict` on inner structures
        """
        return check_loaded_dict(self.my_dict)

    def filter_data(self):  # pragma: no cover
        warnings.warn("Deprecated, use pop errors instead", FutureWarning, stacklevel=2)
        self.pop_errors()

    def pop_errors(self) -> list[tuple[str, dict]]:
        """Remove problematic entries from dict"""
        error_list = []
        for group, up_dkt in list(self.my_dict.items()):
            if not isinstance(up_dkt, (dict, EventedDict)):
                continue
            error_list.extend(
                (f"{group}.{key}", up_dkt.pop(key)) for key, dkt in list(up_dkt.items()) if not check_loaded_dict(dkt)
            )

        return error_list


def check_loaded_dict(dkt) -> bool:
    """
    Recursive check if dict `dkt` or any sub dict contains '__error__' key.

    :param dkt: dict to check
    """
    if not isinstance(dkt, (dict, EventedDict)):
        return True
    if "__error__" in dkt:
        return False
    return all(check_loaded_dict(val) for val in dkt.values())


class BaseModel(PydanticBaseModel):
    class Config:
        extra = "forbid"

    def __getitem__(self, item):
        if item in self.__fields__:
            warnings.warn("Access to attribute by [] is deprecated. Use . instead", FutureWarning, stacklevel=2)
            return getattr(self, item)
        raise KeyError(f"{item} not found in {self.__class__.__name__}")

    def copy(self: PydanticBaseModel, *, validate: bool = True, **kwargs: typing.Any) -> PydanticBaseModel:
        copy_res = super().copy(**kwargs)
        if validate:
            return self.validate(
                dict(
                    copy_res._iter(  # pylint: disable=protected-access
                        to_dict=False, by_alias=False, exclude_unset=True
                    )
                )
            )
        return copy_res


def iterate_names(base_name: str, data_dict, max_length=None) -> typing.Optional[str]:
    if base_name not in data_dict:
        return base_name[:max_length]
    if max_length is not None:
        max_length -= 5
    for i in range(1, 100):
        res_name = f"{base_name[:max_length]} ({i})"
        if res_name not in data_dict:
            return res_name
    return None


def napari_image_repr(image: "Image") -> str:
    return (
        f"<Image of shape: {image.data.shape}, dtype: {image.data.dtype}, "
        f"slice {getattr(image, '_slice_indices', None)}>"
    )


def safe_repr(val):
    from napari.layers import Image

    if isinstance(val, np.ndarray):
        return numpy_repr(val)
    if isinstance(val, Image):
        return napari_image_repr(val)
    return _safe_repr(val)



"""Implementation of PartSeg measurement feature for napari viewer."""

import warnings
from typing import Optional

import numpy as np
from magicgui.widgets import CheckBox, Container, HBox, PushButton, Table, VBox, create_widget
from napari import Viewer
from napari._qt.qthreading import thread_worker
from napari.layers import Image as NapariImage
from napari.layers import Labels as NapariLabels
from napari.qt import create_worker

from PartSeg.plugins import register as register_plugins
from PartSegCore import UNIT_SCALE, Units
from PartSegCore.algorithm_describe_base import base_model_to_algorithm_property
from PartSegCore.analysis.measurement_base import AreaType, Leaf, MeasurementEntry, PerComponent
from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT, MeasurementProfile, MeasurementResult
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


class SimpleMeasurement(Container):
    """Widget that provide access to simple measurement feature from PartSeg ROI Mask."""

    def __init__(self, napari_viewer: Viewer):
        """:param napari_viewer: napari viewer instance."""
        super().__init__(layout="vertical")
        self.viewer = napari_viewer
        self.labels_choice = create_widget(annotation=NapariLabels, label="Labels")
        self.scale_units_select = create_widget(annotation=Units, label="Data units")  # EnumComboBox(Units)
        self.units_select = create_widget(annotation=Units, label="Units")
        self.image_choice = create_widget(annotation=Optional[NapariImage], label="Image", options={})
        self.table = Table()
        self.calculate_btn = PushButton(text="Calculate")
        self.margins = (0, 0, 0, 0)
        self.measurement_result: Optional[MeasurementResult] = None
        self.worker = None

        options_layout = HBox(
            widgets=(
                self.image_choice,
                self.labels_choice,
                self.scale_units_select,
                self.units_select,
                self.calculate_btn,
            )
        )
        self.measurement_layout = VBox()

        bottom_layout = HBox(widgets=(self.measurement_layout, self.table))

        self.insert(0, options_layout)
        self.insert(1, bottom_layout)

        self.image_choice.native.currentIndexChanged.connect(self._refresh_measurements)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)
            self.labels_choice.changed.connect(self._refresh_measurements)
            self.calculate_btn.changed.connect(self._calculate)
        register_plugins()

    def _calculate(self, event=None):
        to_calculate = []
        for chk in self.measurement_layout:
            # noinspection PyTypeChecker
            if chk.value:
                leaf: Leaf = MEASUREMENT_DICT[chk.text].get_starting_leaf()
                to_calculate.append(leaf.replace_(per_component=PerComponent.Yes, area=AreaType.ROI))
        if not to_calculate:  # pragma: no cover
            warnings.warn("No measurement. Select at least one measurement", stacklevel=1)
            return

        profile = MeasurementProfile(
            name="", chosen_fields=[MeasurementEntry(name=x.name, calculation_tree=x) for x in to_calculate]
        )

        data_layer = self.image_choice.value or self.labels_choice.value

        data_ndim = data_layer.data.ndim
        if data_ndim > 4:  # pragma: no cover
            warnings.warn(
                "Not Supported. Currently measurement engine does not support data over 4 dim (TZYX)", stacklevel=1
            )
            return
        data_scale = data_layer.scale[-3:] / UNIT_SCALE[self.scale_units_select.get_value().value]
        image = Image(data_layer.data, spacing=data_scale, axes_order="TZYX"[-data_ndim:])
        worker = _prepare_data(profile, image, self.labels_choice.value.data)
        worker.returned.connect(self._calculate_next)
        worker.errored.connect(self._finished)
        worker.start()
        self.worker = worker
        self.calculate_btn.enabled = False

    def _calculate_next(self, data):
        profile, image, roi_info, segmentation_mask_map = data
        self.measurement_result = MeasurementResult(segmentation_mask_map)
        create_worker(
            profile.calculate_yield,
            _start_thread=True,
            _progress={"total": len(profile.chosen_fields)},
            _connect={"finished": self._finished, "yielded": self._set_result},
            image=image,
            channel_num=0,
            roi=roi_info,
            result_units=self.units_select.get_value(),
            segmentation_mask_map=segmentation_mask_map,
        )

    def _set_result(self, result):
        self.measurement_result[result[0]] = result[1]
        self.table.value = self.measurement_result.to_dataframe()

    def _finished(self):
        self.calculate_btn.enabled = True
        self.worker = None

    def _clean_measurements(self):
        selected = set()
        for _ in range(len(self.measurement_layout)):
            # noinspection PyTypeChecker
            chk = self.measurement_layout.pop(0)
            if chk.value:
                selected.add(chk.text)
        return selected

    def _refresh_measurements(self, event=None):
        has_channel = self.image_choice.value is not None
        selected = self._clean_measurements()
        for val in MEASUREMENT_DICT.values():
            area = val.get_starting_leaf().area
            pc = val.get_starting_leaf().per_component
            if (
                base_model_to_algorithm_property(val.__argument_class__)
                if val.__new_style__
                else val.get_fields()
                or (area is not None and area != AreaType.ROI)
                or (pc is not None and pc != PerComponent.Yes)
            ):
                continue
            text = val.get_name()
            chk = CheckBox(name=text, label=text)
            if val.need_channel() and not has_channel:
                chk.enabled = False
                chk.tooltip = "Need selected image"
            elif text in selected:
                chk.value = True
            self.measurement_layout.insert(-1, chk)

    def reset_choices(self, event=None):
        super().reset_choices(event)
        self._refresh_measurements()


@thread_worker(progress=True)
def _prepare_data(profile: MeasurementProfile, image: Image, labels: np.ndarray):
    roi_info = ROIInfo(labels).fit_to_image(image)
    yield 1
    segmentation_mask_map = profile.get_segmentation_mask_map(image, roi_info, time=0)
    yield 2
    return profile, image, roi_info, segmentation_mask_map



import json

import pytest
from pydantic import BaseModel

from PartSegImage import Channel


def test_channel_from_channel():
    ch1 = Channel(1)
    ch2 = Channel(2)
    ch3 = Channel(ch1)
    assert ch1 == ch3
    assert ch1 != ch2


def test_representation():
    ch1 = Channel(1)
    assert str(ch1) == "2"
    assert repr(ch1) == "<PartSegImage.channel_class.Channel(value=1)>"


def test_as_dict():
    ch1 = Channel(1)
    assert ch1.as_dict() == {"value": 1}
    ch2 = Channel(**ch1.as_dict())
    assert ch1 == ch2


class Model(BaseModel):
    channel: Channel


MODEL_SCHEMA = {
    "title": "Model",
    "type": "object",
    "properties": {
        "channel": {
            "title": "Channel",
            "type": "object",
            "properties": {"value": {"title": "value", "anyOf": [{"type": "string"}, {"type": "integer"}]}},
        }
    },
    "required": ["channel"],
}


@pytest.mark.parametrize("value", [Channel(1), 1, Channel("1"), "1"], ids=lambda x: f"value={x!r}")
def test_as_pydantic_field(value):
    model = Model(channel=value)
    assert isinstance(model.channel, Channel)
    assert model.channel == Channel(value)


def test_json_schema():
    assert json.loads(Model.schema_json()) == MODEL_SCHEMA



import typing
from os.path import basename, dirname, isdir, isfile
from pathlib import Path

from qtpy.QtWidgets import QFileDialog

from PartSegCore.io_utils import LoadBase

if typing.TYPE_CHECKING:  # pragma: no cover
    from PartSeg.common_backend.base_settings import BaseSettings


class LoadProperty(typing.NamedTuple):
    load_location: list[typing.Union[str, Path]]
    selected_filter: str
    load_class: type[LoadBase]


IORegister = typing.Union[dict[str, type(LoadBase)], type(LoadBase), str, list[type(LoadBase)]]


class IOMethodMock:
    __new_style__ = False

    __fields__: typing.ClassVar[dict[str, typing.Any]] = {}

    def __init__(self, name: str):
        self.name = name

    def __call__(self, *args, **kwargs):
        return self

    def get_name(self) -> str:
        return self.name

    def get_name_with_suffix(self) -> str:
        return self.get_name()

    def get_short_name(self) -> str:
        return self.get_name()

    @staticmethod
    def get_default_extension() -> str:
        return ""

    @staticmethod
    def need_segmentation() -> bool:
        return False

    @staticmethod
    def need_mask() -> bool:
        return False

    @staticmethod
    def get_fields() -> list:
        return []

    @staticmethod
    def number_of_files() -> int:
        return 1

    @staticmethod
    def save(*args, **kwargs):
        """For keep compatibility with SaveBase"""


class LoadRegisterFileDialog(QFileDialog):
    def __init__(
        self,
        io_register: IORegister,
        caption,
        parent=None,
    ):
        if isinstance(io_register, str):
            io_register = {io_register: IOMethodMock(io_register)}
        if isinstance(io_register, list):
            io_register = {x.get_name(): x for x in io_register}
        if not isinstance(io_register, typing.MutableMapping):
            io_register = {io_register.get_name(): io_register}
        super().__init__(parent, caption)
        self.io_register = {x.get_name_with_suffix(): x for x in io_register.values()}
        self.setNameFilters(list(self.io_register.keys()))


class CustomLoadDialog(LoadRegisterFileDialog):
    def __init__(
        self,
        load_register: IORegister,
        parent=None,
        caption="Load file",
        history: typing.Optional[list[str]] = None,
    ):
        super().__init__(load_register, caption, parent)
        self.setOption(QFileDialog.Option.DontUseNativeDialog, True)
        self.setFileMode(QFileDialog.FileMode.ExistingFile)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        self.files_list = []
        self.setWindowTitle("Open File")
        if history is not None:
            history = self.history() + history
            self.setHistory(history)

    def accept(self):
        selected_files = [
            x for x in self.selectedFiles() if self.fileMode == QFileDialog.FileMode.Directory or isfile(x)
        ]
        if not selected_files:
            return
        if len(selected_files) == 1 and self.fileMode != QFileDialog.FileMode.Directory and isdir(selected_files[0]):
            super().accept()
            return

        self.files_list.extend(selected_files)
        chosen_class: LoadBase = self.io_register[self.selectedNameFilter()]
        if len(self.files_list) < chosen_class.number_of_files():
            self.setNameFilters([chosen_class.get_name()])
            self.setWindowTitle("Open File for:" + ",".join(basename(x) for x in self.files_list))

            self.selectFile(chosen_class.get_next_file(self.files_list))
        else:
            super().accept()

    def get_result(self) -> LoadProperty:
        chosen_class: type[LoadBase] = self.io_register[self.selectedNameFilter()]
        return LoadProperty(self.files_list, self.selectedNameFilter(), chosen_class)


class PLoadDialog(CustomLoadDialog):
    def __init__(
        self,
        load_register: typing.Union[dict[str, type(LoadBase)], type(LoadBase)],
        *,
        settings: "BaseSettings",
        path: str,
        default_directory: typing.Optional[str] = None,
        filter_path="",
        parent=None,
        caption="Load file",
    ):
        if default_directory is None:
            default_directory = str(Path.home())
        super().__init__(
            load_register=load_register,
            parent=parent,
            caption=caption,
            history=settings.get_path_history(),
        )
        self.settings = settings
        self.path_in_dict = path
        self.filter_path = filter_path
        self.setDirectory(self.settings.get(path, default_directory))
        if self.filter_path:
            self.selectNameFilter(self.settings.get(self.filter_path, ""))

    def accept(self):
        super().accept()
        if self.result() != QFileDialog.DialogCode.Accepted:
            return
        directory = dirname(self.selectedFiles()[0])
        self.settings.add_path_history(directory)
        self.settings.set(self.path_in_dict, directory)
        if self.filter_path:
            self.settings.set(self.filter_path, self.selectedNameFilter())


class SelectDirectoryDialog(QFileDialog):
    def __init__(
        self,
        settings: "BaseSettings",
        settings_path: typing.Union[str, list[str]],
        default_directory: typing.Optional[str] = None,
        parent=None,
    ) -> None:
        super().__init__(parent, "Select directory")
        if default_directory is None:
            default_directory = str(Path.home())
        self.settings = settings
        self.setFileMode(QFileDialog.FileMode.Directory)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        if isinstance(settings_path, list):
            for path_ in reversed(settings_path):
                default_directory = self.settings.get(path_, default_directory)
            self.setDirectory(default_directory)
            self.path_in_dict = settings_path[0]
        else:
            self.setDirectory(self.settings.get(settings_path, default_directory))
            self.path_in_dict = settings_path
        history = self.history() + settings.get_path_history()
        self.setHistory(history)

    def accept(self) -> None:
        super().accept()
        if self.result() != QFileDialog.DialogCode.Accepted:
            return
        directory = self.selectedFiles()[0]
        self.settings.add_path_history(directory)
        self.settings.set(self.path_in_dict, directory)



# pylint: disable=no-self-use
import math
import os.path
import shutil
from glob import glob
from importlib.metadata import version
from io import BytesIO

import numpy as np
import pytest
import tifffile
from packaging.version import parse as parse_version

import PartSegData
from PartSegImage import CziImageReader, GenericImageReader, Image, ObsepImageReader, OifImagReader, TiffImageReader


@pytest.fixture(autouse=True)
def _set_max_workers_czi(monkeypatch):
    # set max workers to 1 to get exception in case of problems
    monkeypatch.setattr("PartSegImage.image_reader.CZI_MAX_WORKERS", 1)


class TestImageClass:
    def test_tiff_image_read(self):
        image = TiffImageReader.read_image(PartSegData.segmentation_mask_default_image)
        assert isinstance(image, Image)
        assert np.all(np.isclose(image.spacing, (7.752248561753867e-08,) * 2))

    def test_tiff_image_read_buffer(self):
        with open(PartSegData.segmentation_mask_default_image, "rb") as f_p:
            buffer = BytesIO(f_p.read())
        image = TiffImageReader.read_image(buffer)
        assert isinstance(image, Image)
        assert np.all(np.isclose(image.spacing, (7.752248561753867e-08,) * 2))

    def test_czi_file_read(self, data_test_dir):
        """Check if czi file is read correctly."""
        image = CziImageReader.read_image(os.path.join(data_test_dir, "test_czi.czi"))
        assert np.count_nonzero(image.get_channel(0))
        assert image.channels == 4
        assert image.layers == 1
        assert image.get_colors() == ["#FFFFFF", "#FF0000", "#00FF00", "#0000FF"]

        assert image.file_path == os.path.join(data_test_dir, "test_czi.czi")

        assert np.all(np.isclose(image.spacing, (7.752248561753867e-08,) * 2))

    @pytest.mark.skipif(
        parse_version(version("czifile")) < parse_version("2019.7.2"),
        reason="There is no patch for czifile before 2019.7.2",
    )
    @pytest.mark.parametrize("file_name", ["test_czi_zstd0.czi", "test_czi_zstd1.czi", "test_czi_zstd1_hilo.czi"])
    def test_czi_file_read_compressed(self, data_test_dir, file_name):
        image = CziImageReader.read_image(os.path.join(data_test_dir, file_name))
        assert np.count_nonzero(image.get_channel(0))
        assert image.channels == 4
        assert image.layers == 1

        assert image.file_path == os.path.join(data_test_dir, file_name)

        assert np.all(np.isclose(image.spacing, (7.752248561753867e-08,) * 2))

    def test_czi_file_read_buffer(self, data_test_dir):
        with open(os.path.join(data_test_dir, "test_czi.czi"), "rb") as f_p:
            buffer = BytesIO(f_p.read())

        image = CziImageReader.read_image(buffer)
        assert np.count_nonzero(image.get_channel(0))
        assert image.channels == 4
        assert image.layers == 1
        assert image.file_path == ""

        assert np.all(np.isclose(image.spacing, (7.752248561753867e-08,) * 2))

    def test_oib_file_read(self, data_test_dir):
        image = OifImagReader.read_image(os.path.join(data_test_dir, "N2A_H2BGFP_dapi_falloidin_cycling1.oib"))
        assert image.channels == 3
        assert image.layers == 6
        assert np.all(np.isclose(image.spacing, (2.1e-07,) + (7.752248561753867e-08,) * 2))

    def test_oif_file_read(self, data_test_dir):
        image = OifImagReader.read_image(os.path.join(data_test_dir, "Image0003_01.oif"))
        assert image.channels == 1
        assert image.layers == 49
        assert np.all(np.isclose(image.spacing, (3.2e-07,) + (5.1e-08,) * 2))

    def test_read_with_mask(self, data_test_dir):
        image = TiffImageReader.read_image(
            os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif"),
            os.path.join(data_test_dir, "stack1_components", "stack1_component1_mask.tif"),
        )
        assert isinstance(image, Image)
        with pytest.raises(ValueError, match="Incompatible shape"):
            TiffImageReader.read_image(
                os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif"),
                os.path.join(data_test_dir, "stack1_components", "stack1_component2_mask.tif"),
            )

    def test_lsm_read(self, data_test_dir):
        image1 = TiffImageReader.read_image(os.path.join(data_test_dir, "test_lsm.lsm"))
        image2 = TiffImageReader.read_image(os.path.join(data_test_dir, "test_lsm.tif"))
        data = np.moveaxis(np.load(os.path.join(data_test_dir, "test_lsm.npy")), -1, 0)
        assert np.all(image1.get_data() == data)
        assert np.all(image2.get_data() == data)
        assert np.all(image1.get_data() == image2.get_data())

    def test_ome_read(self, data_test_dir):  # error in tifffile
        image1 = TiffImageReader.read_image(os.path.join(data_test_dir, "test_lsm2.tif"))
        image2 = TiffImageReader.read_image(os.path.join(data_test_dir, "test_lsm.tif"))
        data = np.moveaxis(np.load(os.path.join(data_test_dir, "test_lsm.npy")), -1, 0)
        assert np.all(image1.get_data() == data)
        assert np.all(image2.get_data() == data)
        assert np.all(image1.get_data() == image2.get_data())

    def test_generic_reader(self, data_test_dir):
        GenericImageReader.read_image(
            os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif"),
            os.path.join(data_test_dir, "stack1_components", "stack1_component1_mask.tif"),
        )
        GenericImageReader.read_image(os.path.join(data_test_dir, "test_czi.czi"))
        GenericImageReader.read_image(os.path.join(data_test_dir, "test_lsm2.tif"))
        GenericImageReader.read_image(os.path.join(data_test_dir, "test_lsm.tif"))
        GenericImageReader.read_image(os.path.join(data_test_dir, "Image0003_01.oif"))
        GenericImageReader.read_image(os.path.join(data_test_dir, "N2A_H2BGFP_dapi_falloidin_cycling1.oib"))

    def test_generic_reader_from_buffer(self, data_test_dir):
        file_path = os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif")
        with open(file_path, "rb") as f_p:
            buffer = BytesIO(f_p.read())
        GenericImageReader().read(buffer)
        with open(file_path, "rb") as f_p:
            buffer = BytesIO(f_p.read())
        GenericImageReader().read(buffer, ext=".tif")
        with pytest.raises(NotImplementedError, match="Oif format is not supported"):
            GenericImageReader().read(buffer, ext=".oif")
        with pytest.raises(NotImplementedError, match="Obsep format is not supported"):
            GenericImageReader().read(buffer, ext=".obsep")

    def test_decode_int(self):
        assert TiffImageReader.decode_int(0) == [0, 0, 0, 0]
        assert TiffImageReader.decode_int(15) == [0, 0, 0, 15]
        assert TiffImageReader.decode_int(3 + 7 * 256 + 11 * 256**2 + 13 * 256**3) == [13, 11, 7, 3]

    def test_set_spacing(self):
        reader = TiffImageReader()
        reader.set_default_spacing((11, 12, 13))
        assert reader.default_spacing == (11, 12, 13)
        reader.set_default_spacing((5, 7))
        assert reader.default_spacing == (10**-6, 5, 7)

    def test_obsep_read(self, data_test_dir):
        image = ObsepImageReader.read_image(os.path.join(data_test_dir, "obsep", "test.obsep"))
        assert image.channels == 2
        assert np.allclose(image.spacing, (500 * 10**-9, 64 * 10**-9, 64 * 10**-9))
        assert image.channel_names == ["channel 1", "channel 2"]

    def test_obsep_deconv_read(self, data_test_dir, tmp_path):
        for el in glob(os.path.join(data_test_dir, "obsep", "*")):
            shutil.copy(os.path.join(data_test_dir, "obsep", el), tmp_path)
        image = GenericImageReader.read_image(tmp_path / "test.obsep")
        assert image.channels == 2
        assert np.allclose(image.spacing, (500 * 10**-9, 64 * 10**-9, 64 * 10**-9))
        assert image.channel_names == ["channel 1", "channel 2"]
        shutil.copy(tmp_path / "Cy5.TIF", tmp_path / "Cy5_decon2.TIF")
        image = GenericImageReader.read_image(tmp_path / "test.obsep")
        assert image.channels == 2
        shutil.copy(tmp_path / "Cy5.TIF", tmp_path / "Cy5_deconv.TIF")
        image = GenericImageReader.read_image(tmp_path / "test.obsep")
        assert image.channels == 3

    def test_double_axes_in_dim_read(self, data_test_dir):
        image = GenericImageReader.read_image(os.path.join(data_test_dir, "double_q_in_axes.tif"))
        assert image.layers == 360
        assert image.channels == 1
        assert image.stack_pos == 1
        assert image.plane_shape == (360, 32)


class CustomImage(Image):
    axis_order = "TCXYZ"


class CustomTiffReader(TiffImageReader):
    image_class = CustomImage


def test_change_class(data_test_dir):
    img = CustomTiffReader.read_image(os.path.join(data_test_dir, "test_lsm.tif"))
    assert isinstance(img, CustomImage)
    assert img.plane_shape == (1024, 1024)
    assert img.layers == 6
    assert img.channels == 3
    assert img.stack_pos == 3


def test_xml2dict():
    sample_text = """
    <level1>
        <level2>3.5322</level2>
    </level1>
    """
    data = tifffile.xml2dict(sample_text)
    assert math.isclose(data["level1"]["level2"], 3.5322)
